import { default as chloride } from "chloride";
import { Buffer } from "buffer";
import crypto from "crypto";
import { concatBytes, identifyBufOrBytes, stringToBytes, } from "../util/bytes.js";
import { bufferToBytes, bytesToBuffer, stringToBuffer, } from "../util/buffers.js";
//--------------------------------------------------
import { Logger, LogLevel, setLogLevel } from "../util/log.js";
import { UpdatableHash } from "./updatable_hash.js";
let logger = new Logger("crypto-driver-chloride", "cyan");
setLogLevel("crypto-driver-chloride", LogLevel.Info);
//================================================================================
/**
 * A version of the ICryptoDriver interface backed by Chloride.
 * Faster than noble.
 * Works in Node and the browser (with some polyfilling on your part).
 */
export const CryptoDriverChloride = class {
    static sha256(input) {
        if (typeof input === "string")
            input = stringToBuffer(input);
        if (identifyBufOrBytes(input) === "bytes")
            input = bytesToBuffer(input);
        const resultBuf = chloride.crypto_hash_sha256(input);
        return Promise.resolve(bufferToBytes(resultBuf));
    }
    static updatableSha256() {
        return new UpdatableHash({
            hash: crypto.createHash("sha256"),
            update: (hash, data) => hash.update(data),
            digest: (hash) => {
                const digest = hash.digest();
                if (typeof digest === "string") {
                    return stringToBytes(digest);
                }
                return bufferToBytes(digest);
            },
        });
    }
    static generateKeypairBytes(seed) {
        // If provided, the seed is used as the secret key.
        // If omitted, a random secret key is generated.
        logger.debug("generateKeypairBytes");
        let seedBuf = seed === undefined ? undefined : bytesToBuffer(seed);
        if (!seedBuf) {
            seedBuf = Buffer.alloc(32);
            chloride.randombytes(seedBuf);
        }
        const keys = chloride.crypto_sign_seed_keypair(seedBuf);
        return Promise.resolve({
            //curve: 'ed25519',
            pubkey: bufferToBytes(keys.publicKey),
            // so that this works with either sodium or libsodium-wrappers (in browser):
            secret: bufferToBytes((keys.secretKey).slice(0, 32)),
        });
    }
    static sign(keypairBytes, msg) {
        logger.debug("sign");
        const secretBuf = bytesToBuffer(concatBytes(keypairBytes.secret, keypairBytes.pubkey));
        if (typeof msg === "string")
            msg = stringToBuffer(msg);
        if (msg instanceof Uint8Array)
            msg = bytesToBuffer(msg);
        return Promise.resolve(bufferToBytes(
        // this returns a Buffer
        chloride.crypto_sign_detached(msg, secretBuf)));
    }
    static verify(publicKey, sig, msg) {
        logger.debug("verify");
        try {
            if (typeof msg === "string")
                msg = stringToBuffer(msg);
            if (msg instanceof Uint8Array)
                msg = bytesToBuffer(msg);
            return Promise.resolve(chloride.crypto_sign_verify_detached(bytesToBuffer(sig), msg, publicKey));
        }
        catch {
            /* istanbul ignore next */
            return Promise.resolve(false);
        }
    }
};
