import { stringToBytes } from "../util/bytes.js";
import { ed, sha256_uint8array } from "../../deps.js";
const { createHash } = sha256_uint8array;
//--------------------------------------------------
import { Logger } from "../util/log.js";
import { UpdatableHash } from "./updatable_hash.js";
const logger = new Logger("crypto-driver-noble", "cyan");
//================================================================================
/**
 * A version of the ICryptoDriver interface backed by noble/ed25519.
 * The slowest crypto driver available, but works everywhere.
 */
export const CryptoDriverNoble = class {
    static sha256(input) {
        if (typeof input === "string") {
            return Promise.resolve(createHash("sha256").update(input, "utf-8").digest());
        }
        else {
            return Promise.resolve(createHash("sha256").update(input).digest());
        }
    }
    static updatableSha256() {
        return new UpdatableHash({
            hash: createHash("sha256"),
            update: (hash, data) => hash.update(data),
            digest: (hash) => hash.digest(),
        });
    }
    static async generateKeypairBytes() {
        logger.debug("generateKeypairBytes");
        const secret = ed.utils.randomPrivateKey();
        const pubkey = await ed.getPublicKey(secret);
        return {
            pubkey,
            secret,
        };
    }
    static sign(keypairBytes, msg) {
        logger.debug("sign");
        if (typeof msg === "string")
            msg = stringToBytes(msg);
        return ed.sign(msg, keypairBytes.secret);
    }
    static async verify(publicKey, sig, msg) {
        logger.debug("verify");
        try {
            if (typeof msg === "string")
                msg = stringToBytes(msg);
            const result = await ed.verify(sig, msg, publicKey);
            return result;
        }
        catch {
            return false;
        }
    }
};
