/**
 * [Earthstar](https://earthstar-project.org) is a small and resilient distributed storage protocol designed with a strong focus on simplicity and versatility, with the social realities of peer-to-peer computing kept in mind.
 *
 * This is a reference implementation written in Typescript. You can use it to add Earthstar functionality to applications running on servers, browsers, the command line, or anywhere else JavaScript can be run.
 *
 * ### Example usage
 *
 * ```ts
 * import { Replica, ReplicaDriverMemory, Crypto, Peer } from "earthstar";
 *
 * const shareKeypair = await Crypto.generateShareKeypair("gardening");
 *
 * const replica = new Replica({
 * 	 driver: ReplicaDriverMemory(shareKeypair.shareAddress),
 * 	 shareSecret: shareKeypair.secret,
 * });
 *
 * const authorKeypair = await Crypto.generateAuthorKeypair("suzy");
 *
 * await replica.set(authorKeypair, {
 * 	 path: "/my-note",
 * 	 text: "Saw seven magpies today",
 * });
 *
 * const allDocs = await replica.getAllDocs();
 *
 * const peer = new Peer();
 *
 * peer.addReplica(replica);
 *
 * peer.sync("https://my.server")
 * ```
 *
 * This module also exposes server APIs for for building always-online peers. The below example reads some on-disk JSON to initiate some share replicas, and stores their data using the filesystem.
 *
 * ```ts
 * import {
 * ExtensionKnownShares,
 * ExtensionSyncWeb,
 * Server,
 * } from "https://deno.land/x/earthstar/mod.ts";
 *
 * const server = new Server([
 * new ExtensionKnownShares({
 * 	 knownSharesPath: "./known_shares.json",
 * 	 onCreateReplica: (shareAddress) => {
 * 		 return new Earthstar.Replica({
 * 			 driver: new ReplicaDriverFs(shareAddress, "./share_data"),
 * 		 });
 * 	 },
 * }),
 * new ExtensionSyncWebsocket(),
 * ]);
 *
 * @module
 */
export * from "../core-validators/addresses.js";
export * from "../core-validators/characters.js";
export * from "../core-validators/checkers.js";
export * from "../crypto/base32.js";
export * from "../crypto/crypto-driver-noble.js";
export * from "../crypto/crypto-types.js";
export * from "../crypto/crypto.js";
export * from "../crypto/global-crypto-driver.js";
export * from "../crypto/keypair.js";
export * from "../formats/format_es4.js";
export * from "../formats/format_es5.js";
export * from "../formats/util.js";
export * from "../formats/format_types.js";
export * from "../syncer/syncer.js";
export * from "../syncer/partner_local.js";
export * from "../syncer/partner_web_client.js";
export * from "../syncer/syncer_types.js";
export * from "../peer/peer-types.js";
export * from "../peer/peer.js";
export * from "../query/query-types.js";
export * from "../query/query.js";
export * from "../query/query-helpers.js";
export * from "../replica/compare.js";
export * from "../replica/replica.js";
export * from "../replica/replica_cache.js";
export * from "../replica/replica-types.js";
export * from "../replica/multiformat_replica.js";
export * from "../replica/util-types.js";
export * from "../replica/doc_drivers/memory.js";
export * from "../replica/attachment_drivers/memory.js";
export * from "../replica/driver_memory.js";
export * from "../util/bytes.js";
export * from "../util/doc-types.js";
export * from "../util/errors.js";
export * from "../util/invite.js";
export * from "../util/log.js";
export * from "../util/misc.js";
export * from "../util/shared_settings.js";
