import * as dntShim from "../../../_dnt.shims.js";
import { Crypto } from "../../crypto/crypto.js";
import { ReplicaIsClosedError, ValidationError } from "../../util/errors.js";
import { streamToBytes } from "../../util/streams.js";
/** An attachment driver which persists attachments in memory.
 * Works everywhere.
 */
export class AttachmentDriverMemory {
    constructor() {
        Object.defineProperty(this, "stagingMap", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Map()
        });
        Object.defineProperty(this, "attachmentMap", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Map()
        });
        Object.defineProperty(this, "closed", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: false
        });
    }
    getKey(formatName, attachmentHash) {
        return `${formatName}___${attachmentHash}`;
    }
    getAttachment(formatName, attachmentHash) {
        if (this.closed)
            throw new ReplicaIsClosedError();
        const key = this.getKey(formatName, attachmentHash);
        const attachment = this.attachmentMap.get(key);
        if (!attachment) {
            return Promise.resolve(undefined);
        }
        return Promise.resolve({
            bytes: async () => new Uint8Array(await attachment.arrayBuffer()),
            stream: () => Promise.resolve(
            // Need to do this for Node's sake.
            attachment.stream()),
        });
    }
    async stage(formatName, attachment) {
        if (this.closed)
            throw new ReplicaIsClosedError();
        const bytes = attachment instanceof Uint8Array
            ? attachment
            : await streamToBytes(attachment);
        const hash = await Crypto.sha256base32(bytes);
        const newAttachment = new dntShim.Blob([bytes]);
        const key = this.getKey(formatName, hash);
        this.stagingMap.set(key, newAttachment);
        return Promise.resolve({
            hash,
            size: bytes.byteLength,
            commit: () => {
                this.attachmentMap.set(key, newAttachment);
                this.stagingMap.delete(key);
                return Promise.resolve();
            },
            reject: () => {
                this.stagingMap.delete(key);
                return Promise.resolve();
            },
        });
    }
    erase(formatName, attachmentHash) {
        if (this.closed)
            throw new ReplicaIsClosedError();
        const key = this.getKey(formatName, attachmentHash);
        if (this.attachmentMap.has(key)) {
            this.attachmentMap.delete(key);
            return Promise.resolve(true);
        }
        return Promise.resolve(new ValidationError("No attachment with that signature found."));
    }
    wipe() {
        if (this.closed)
            throw new ReplicaIsClosedError();
        this.attachmentMap.clear();
        return Promise.resolve();
    }
    async filter(hashes) {
        if (this.closed)
            throw new ReplicaIsClosedError();
        const erasedAttachments = [];
        for (const key of this.attachmentMap.keys()) {
            const [format, hash] = key.split("___");
            const hashesToKeep = hashes[format];
            if (hashesToKeep && !hashesToKeep.has(hash)) {
                const result = await this.erase(format, hash);
                if (result) {
                    erasedAttachments.push({ format, hash });
                }
            }
        }
        return erasedAttachments;
    }
    clearStaging() {
        if (this.closed)
            throw new ReplicaIsClosedError();
        this.stagingMap.clear();
        return Promise.resolve();
    }
    isClosed() {
        return this.closed;
    }
    async close(erase) {
        if (this.closed)
            throw new ReplicaIsClosedError();
        if (erase) {
            await this.wipe();
        }
        this.closed = true;
        return;
    }
}
