import { ReplicaIsClosedError } from "../../util/errors.js";
import { DocDriverMemory } from "./memory.js";
//--------------------------------------------------
import { Logger } from "../../util/log.js";
let logger = new Logger("storage driver localStorage", "gold");
function isSerializedDriverDocs(value) {
    // check if data we've loaded from localStorage is actually in the format we expect
    if (typeof value !== "object") {
        return false;
    }
    return ("byPathAndAuthor" in value && "byPathNewestFirst" in value);
}
/** A replica driver which perists to LocalStorage, which stores a maximum of five megabytes per domain. If you're storing multiple shares, this limit will be divided among all their replicas.
 * Works in browsers and Deno.
 */
export class DocDriverLocalStorage extends DocDriverMemory {
    /**
     * @param share - The address of the share the replica belongs to.
     * @param key - An optional key you can use to differentiate storage for the same share on the same device.
     */
    constructor(share, key) {
        super(share);
        Object.defineProperty(this, "_localStorageKeyConfig", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_localStorageKeyDocs", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        logger.debug("constructor");
        // each config item starts with this prefix and gets its own entry in localstorage
        this._localStorageKeyConfig = `earthstar:config:${share}${key ? `:${key}` : ""}`;
        // but all docs are stored together inside this one item, as a giant JSON object
        this._localStorageKeyDocs = `earthstar:documents:${share}${key ? `:${key}` : ""}`;
        const existingData = localStorage.getItem(this._localStorageKeyDocs);
        if (existingData !== null) {
            logger.debug("...constructor: loading data from localStorage");
            const parsed = JSON.parse(existingData);
            if (!isSerializedDriverDocs(parsed)) {
                console.warn(`localStorage data could not be parsed for share ${share}`);
                return;
            }
            this.docByPathAndAuthor = new Map(Object.entries(parsed.byPathAndAuthor));
            this.docsByPathNewestFirst = new Map(Object.entries(parsed.byPathNewestFirst));
            this.latestDocsByPath = new Map(Object.entries(parsed.latestDocsByPath));
            const localIndexes = Array.from(this.docByPathAndAuthor.values()).map((doc) => doc._localIndex);
            this._maxLocalIndex = Math.max(...localIndexes);
        }
        else {
            logger.debug("...constructor: there was no existing data in localStorage");
        }
        logger.debug("...constructor is done.");
    }
    //--------------------------------------------------
    // LIFECYCLE
    // isClosed(): inherited
    close(erase) {
        logger.debug("close");
        if (this._isClosed)
            throw new ReplicaIsClosedError();
        if (erase) {
            logger.debug("...close: and erase");
            this._configKv = {};
            this._maxLocalIndex = -1;
            this.docsByPathNewestFirst.clear();
            this.docByPathAndAuthor.clear();
            logger.debug("...close: erasing localStorage");
            localStorage.removeItem(this._localStorageKeyDocs);
            for (const key of this._listConfigKeysSync()) {
                this._deleteConfigSync(key);
            }
            logger.debug("...close: erasing is done");
        }
        this._isClosed = true;
        logger.debug("...close is done.");
        return Promise.resolve();
    }
    //--------------------------------------------------
    // CONFIG
    // synchronous versions for internal use
    _getConfigSync(key) {
        if (this._isClosed)
            throw new ReplicaIsClosedError();
        key = `${this._localStorageKeyConfig}:${key}`;
        let result = localStorage.getItem(key);
        return result === null ? undefined : result;
    }
    _setConfigSync(key, value) {
        if (this._isClosed)
            throw new ReplicaIsClosedError();
        key = `${this._localStorageKeyConfig}:${key}`;
        localStorage.setItem(key, value);
    }
    _listConfigKeysSync() {
        if (this._isClosed)
            throw new ReplicaIsClosedError();
        let keys = Object.keys(localStorage)
            .filter((key) => key.startsWith(this._localStorageKeyConfig + ":"))
            .map((key) => key.slice(this._localStorageKeyConfig.length + 1));
        keys.sort();
        return keys;
    }
    _deleteConfigSync(key) {
        if (this._isClosed)
            throw new ReplicaIsClosedError();
        let hadIt = this._getConfigSync(key);
        key = `${this._localStorageKeyConfig}:${key}`;
        localStorage.removeItem(key);
        return hadIt !== undefined;
    }
    // async versions to match the IStorageDriverAsync interface
    async getConfig(key) {
        return this._getConfigSync(key);
    }
    async setConfig(key, value) {
        return this._setConfigSync(key, value);
    }
    async listConfigKeys() {
        return this._listConfigKeysSync();
    }
    async deleteConfig(key) {
        return this._deleteConfigSync(key);
    }
    //--------------------------------------------------
    // GET
    // getMaxLocalIndex(): inherited
    // queryDocs(query: Query): inherited
    //--------------------------------------------------
    // SET
    async upsert(doc) {
        if (this._isClosed)
            throw new ReplicaIsClosedError();
        const upsertedDoc = await super.upsert(doc);
        // After every upsert, for now, we save everything
        // to localStorage as a single giant JSON attachment.
        // TODO: debounce this, only do it every 1 second or something
        const docsToBeSerialised = {
            byPathAndAuthor: Object.fromEntries(this.docByPathAndAuthor),
            byPathNewestFirst: Object.fromEntries(this.docsByPathNewestFirst),
            latestDocsByPath: Object.fromEntries(this.latestDocsByPath),
        };
        localStorage.setItem(this._localStorageKeyDocs, JSON.stringify(docsToBeSerialised));
        return upsertedDoc;
    }
}
