import * as dntShim from "../../../_dnt.shims.js";
import { isErr } from "../../util/errors.js";
import { contentType } from "../../../deps/deno.land/std@0.167.0/media_types/mod.js";
import { extname } from "../../../deps/deno.land/std@0.154.0/path/mod.js";
/** A server extension for exposing the contents of shares, so that you can request documents by their path and have them served over HTTP. Can be used to create wikis, websites, image galleries, and more. */
export class ExtensionServeContent {
    constructor(opts) {
        Object.defineProperty(this, "peer", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: null
        });
        Object.defineProperty(this, "path", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: "/"
        });
        Object.defineProperty(this, "replica", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: null
        });
        Object.defineProperty(this, "sourceShare", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "indexPath", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "allowedOrigins", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: []
        });
        if (opts.path) {
            this.path = opts.path;
        }
        this.sourceShare = opts.sourceShare;
        this.indexPath = opts.indexPath;
        if (opts.allowedOrigins) {
            this.allowedOrigins = opts.allowedOrigins;
        }
    }
    register(peer) {
        this.peer = peer;
        const replica = this.peer.getReplica(this.sourceShare);
        if (!replica) {
            throw new Error(`No replica belonging to share ${this.sourceShare} was found!`);
        }
        this.replica = replica;
        return Promise.resolve();
    }
    async handler(req) {
        const pathPattern = new dntShim.URLPattern({
            pathname: `${this.path}*`,
        });
        const pathPatternResult = pathPattern.exec(req.url);
        if (this.replica && pathPatternResult && req.method === "GET") {
            const pathToGet = pathPatternResult.pathname.groups[0];
            if (pathToGet === "" && this.indexPath) {
                const doc = await this.replica.getLatestDocAtPath(this.indexPath);
                if (doc) {
                    const attachment = await this.replica.getAttachment(doc);
                    if (attachment && !isErr(attachment)) {
                        return new dntShim.Response(await attachment.stream(), {
                            headers: {
                                status: "200",
                                "content-type": getContentType(doc.path),
                                "access-control-allow-origin": `localhost ${this.allowedOrigins.join(", ")}`,
                            },
                        });
                    }
                    return new dntShim.Response(doc.text, {
                        headers: {
                            status: "200",
                            "content-type": getContentType(doc.path),
                            "access-control-allow-origin": `localhost ${this.allowedOrigins.join(", ")}`,
                        },
                    });
                }
            }
            const maybeDocument = await this.replica.getLatestDocAtPath(`/${pathToGet}`);
            if (!maybeDocument) {
                return new dntShim.Response("Not found", {
                    headers: {
                        status: "404",
                    },
                });
            }
            const attachment = await this.replica.getAttachment(maybeDocument);
            if (attachment && !isErr(attachment)) {
                return new dntShim.Response(await attachment.stream(), {
                    headers: {
                        status: "200",
                        "content-type": getContentType(maybeDocument.path),
                        "access-control-allow-origin": `localhost ${this.allowedOrigins.join(", ")}`,
                    },
                });
            }
            if (attachment === undefined) {
                return new dntShim.Response(`Not found: ${maybeDocument.text}, ${maybeDocument.attachmentSize}`, {
                    headers: {
                        status: "404",
                        "content-type": getContentType(maybeDocument.path),
                        "access-control-allow-origin": `localhost ${this.allowedOrigins.join(", ")}`,
                    },
                });
            }
            return new dntShim.Response(maybeDocument.text, {
                headers: {
                    status: "200",
                    "content-type": getContentType(maybeDocument.path),
                    "access-control-allow-origin": `localhost ${this.allowedOrigins.join(", ")}`,
                },
            });
        }
        return Promise.resolve(null);
    }
}
function getContentType(path) {
    const extension = extname(path);
    return contentType(extension) || "text/plain";
}
