import * as dntShim from "../../../_dnt.shims.js";
import { deferred } from "../../../deps/deno.land/std@0.138.0/async/deferred.js";
import { PartnerWebClient } from "../../syncer/partner_web_client.js";
import { randomId } from "../../util/misc.js";
import { WebSocketServer } from "ws";
/** An extension which enables synchronisation over the web via HTTP. */
export class ExtensionSyncWeb {
    constructor(opts) {
        Object.defineProperty(this, "path", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: ""
        });
        Object.defineProperty(this, "syncers", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Map()
        });
        Object.defineProperty(this, "peer", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: deferred()
        });
        Object.defineProperty(this, "formats", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "wss", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "server", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        if (opts?.path) {
            this.path = opts.path;
        }
        if (opts?.formats) {
            this.formats = opts.formats;
        }
        this.server = opts.server;
        this.wss = new WebSocketServer({ noServer: true });
    }
    register(peer) {
        this.peer.resolve(peer);
        const transferPattern = new dntShim.URLPattern({
            pathname: `${this.path}/:syncerId/:kind/:shareAddress/:formatName/:author/:path*`,
        });
        const initiatePattern = new dntShim.URLPattern({
            pathname: `${this.path}/:mode`,
        });
        this.server.on("upgrade", (req, socket, head) => {
            const reqUrl = `http://0.0.0.0${req.url}`;
            if (transferPattern.test(reqUrl) || initiatePattern.test(reqUrl)) {
                this.wss.handleUpgrade(req, socket, head, (ws) => {
                    // @ts-ignore
                    this.wss.emit("connection", ws, req);
                });
            }
        });
        this.wss.on("connection", async (socket, req) => {
            const reqUrl = `http://0.0.0.0${req.url}`;
            const transferMatch = transferPattern.exec(reqUrl);
            if (transferMatch) {
                const { syncerId, shareAddress, formatName, path, author, kind } = transferMatch.pathname.groups;
                const syncer = this.syncers.get(syncerId);
                if (!syncer) {
                    return;
                }
                await syncer.handleTransferRequest({
                    shareAddress,
                    formatName,
                    path: `/${path}`,
                    author,
                    kind: kind,
                    source: socket,
                });
                return;
            }
            const initiateMatch = initiatePattern.exec(reqUrl);
            if (initiateMatch) {
                const { mode } = initiateMatch.pathname.groups;
                if (mode !== "once" && mode !== "continuous") {
                    return;
                }
                const peer = await this.peer;
                const partner = new PartnerWebClient({
                    socket,
                    appetite: mode === "once" ? "once" : "continuous",
                });
                const description = `Client ${randomId()}`;
                const newSyncer = peer.addSyncPartner(partner, description, this.formats);
                console.log(`${description}: started sync`);
                newSyncer.isDone().then(() => {
                    console.log(`${description}: completed sync`);
                }).catch((err) => {
                    console.error(`Syncer ${newSyncer.id}: cancelled`, err);
                }).finally(() => {
                    console.log(`${description}: removed`);
                    this.syncers.delete(newSyncer.id);
                });
                this.syncers.set(newSyncer.id, newSyncer);
            }
        });
        return Promise.resolve();
    }
    async handler(req) {
        return Promise.resolve(null);
    }
}
