import * as dntShim from "../../_dnt.shims.js";
import { Buffer } from "node:buffer";
import { ServerCore } from "./server_core.js";
/**
 * An extensible server able to synchronise with other peers.
 *
 * A server's functionality can be extended using extensions of type `IServerExtension`.
 */
export class Server {
    /**
     * Create a new server with an array of extensions.
     * @param extensions - The extensions used by the server. Extensions will be registered in the order you provide them in, as one extension may depend on the actions of another. For example, the `ExtensionServeContent` may rely on a replica created by `ExtensionKnownShares`.
     */
    constructor(extensions, opts) {
        Object.defineProperty(this, "core", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "server", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        this.core = new ServerCore(extensions);
        this.server = opts.server;
        this.server.on("request", async (req, res) => {
            let data = undefined;
            if (req.method === "POST") {
                const buffers = [];
                for await (const chunk of req) {
                    buffers.push(chunk);
                }
                data = Buffer.concat(buffers).toString();
            }
            const headers = new dntShim.Headers();
            for (const key in req.headers) {
                if (req.headers[key])
                    headers.append(key, req.headers[key]);
            }
            // Need the hostname here so the URL plays nice with Node's URL class.
            const url = `http://0.0.0.0${req.url}`;
            const request = new dntShim.Request(url, {
                method: req.method,
                headers,
                body: data,
            });
            const response = await this.core.handler(request);
            // Headers
            for (const [key, value] of response.headers) {
                res.setHeader(key, value);
            }
            // Status
            res.statusCode = response.status;
            // Body
            // TODO: Handle streaming responses.
            if (response.body) {
                res.end(response.body);
            }
            res.end();
        });
        this.server.listen(opts?.port);
    }
    close() {
        this.server.close();
    }
}
