import * as dntShim from "../../_dnt.shims.js";
import { AsyncQueue, deferred } from "../../deps.js";
import { websocketReadable, websocketWritable, } from "../streams/stream_utils.js";
import { EarthstarError, NotSupportedError } from "../util/errors.js";
/** A syncing partner to be used with servers reachable via the internet.
 * Works everywhere.
 */
export class PartnerWebServer {
    constructor(opts) {
        Object.defineProperty(this, "syncAppetite", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "concurrentTransfers", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: 16
        });
        Object.defineProperty(this, "payloadThreshold", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: 8
        });
        Object.defineProperty(this, "rangeDivision", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: 8
        });
        Object.defineProperty(this, "isSecure", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "wsUrl", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "socket", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "incomingQueue", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new AsyncQueue()
        });
        Object.defineProperty(this, "socketIsReady", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: deferred()
        });
        this.syncAppetite = opts.appetite;
        // Check if it's a URL of some kind.
        const url = new URL(opts.url);
        // Check if it's a web syncer
        const hostAndPath = `${url.host}${url.pathname === "/" ? "" : url.pathname}`;
        this.isSecure = url.protocol === "https:" ||
            url.protocol === "wss:";
        this.wsUrl = `${this.isSecure ? "wss://" : "ws://"}${hostAndPath}/'`;
        const urlWithMode = new URL(opts.appetite, this.wsUrl);
        this.socket = new dntShim.WebSocket(urlWithMode.toString());
        this.socket.onopen = () => {
            this.socketIsReady.resolve();
        };
        this.socket.binaryType = "arraybuffer";
        this.socket.onmessage = (event) => {
            // Casting to string due to Node's incorrect websocket types
            this.incomingQueue.push(JSON.parse(event.data));
        };
        this.socket.onclose = () => {
            this.incomingQueue.close();
        };
        this.socket.onerror = (event) => {
            if ("error" in event) {
                this.incomingQueue.close({
                    withError: event.error,
                });
                return;
            }
            this.incomingQueue.close({
                withError: new EarthstarError("Websocket error."),
            });
        };
    }
    async sendEvent(event) {
        await this.socketIsReady;
        if (this.socket.readyState === this.socket.CLOSED ||
            this.socket.readyState === this.socket.CLOSING) {
            return;
        }
        return this.socket.send(JSON.stringify(event));
    }
    getEvents() {
        return this.incomingQueue;
    }
    closeConnection() {
        this.socket.close();
        return Promise.resolve();
    }
    getDownload(opts) {
        // create a new url with the share, path, and syncer ID embedded
        const url = new URL(`${opts.syncerId}/download/${opts.shareAddress}/${opts.doc.format}/${opts.doc.author}${opts.doc.path}`, this.wsUrl);
        const readable = websocketReadable(url.toString(), (event) => {
            if (event.data instanceof ArrayBuffer) {
                const bytes = new Uint8Array(event.data);
                return bytes;
            }
            return null;
        });
        return Promise.resolve(readable);
    }
    handleUploadRequest(opts) {
        const url = new URL(`${opts.syncerId}/upload/${opts.shareAddress}/${opts.doc.format}/${opts.doc.author}${opts.doc.path}`, this.wsUrl);
        const writable = websocketWritable(url.toString(), (outgoing) => outgoing);
        return Promise.resolve(writable);
    }
    handleTransferRequest(_source, _kind) {
        // We don't expect any external requests.
        return Promise.resolve(new NotSupportedError("SyncDriverWebClient does not support external transfer requests."));
    }
}
