import { BlockingBus } from "../streams/stream_utils.js";
import { PromiseEnroller } from "./promise_enroller.js";
export class TransferQueue {
    constructor(activeLimit) {
        Object.defineProperty(this, "waiting", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: []
        });
        Object.defineProperty(this, "active", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Set()
        });
        Object.defineProperty(this, "failed", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Set()
        });
        Object.defineProperty(this, "completed", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Set()
        });
        Object.defineProperty(this, "activeLimit", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "transfersRequestedByUsEnroller", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new PromiseEnroller(true)
        });
        // This status is going to be modified a LOT so it's better to mutate than recreate from scratch.
        Object.defineProperty(this, "reports", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: {}
        });
        Object.defineProperty(this, "reportBus", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new BlockingBus()
        });
        this.activeLimit = activeLimit;
    }
    async activate(transfer) {
        this.active.add(transfer);
        transfer.isDone().then(() => {
            this.completed.add(transfer);
        }).catch(() => {
            this.failed.add(transfer);
        }).finally(() => {
            this.active.delete(transfer);
            this.admitNext();
        });
        await transfer.start();
    }
    queue(transfer) {
        this.waiting.push(transfer);
    }
    admitNext() {
        if (this.waiting.length === 0) {
            return;
        }
        if (this.active.size >= this.activeLimit) {
            return;
        }
        const first = this.waiting.shift();
        if (first) {
            this.activate(first);
        }
    }
    async addTransfer(transfer) {
        transfer.onProgress(() => {
            this.updateTransferStatus(transfer);
        });
        if (transfer.requester === "us") {
            this.transfersRequestedByUsEnroller.enrol(transfer.isDone());
        }
        if (this.active.size < this.activeLimit) {
            await this.activate(transfer);
        }
        else {
            this.queue(transfer);
        }
    }
    gotAllTransfersRequestedByUs() {
        this.transfersRequestedByUsEnroller.seal();
    }
    cancel() {
        this.transfersRequestedByUsEnroller.seal();
        for (const transfer of this.active) {
            transfer.abort();
        }
    }
    updateTransferStatus(transfer) {
        const shareReports = this.reports[transfer.share];
        if (!shareReports) {
            this.reports[transfer.share] = {};
        }
        this.reports[transfer.share][transfer.hash + transfer.kind] = {
            author: transfer.doc.author,
            path: transfer.doc.path,
            format: transfer.doc.format,
            hash: transfer.hash,
            kind: transfer.kind,
            status: transfer.status,
            bytesLoaded: transfer.loaded,
            totalBytes: transfer.expectedSize,
        };
        this.reportBus.send(this.getReport());
    }
    getReport() {
        const report = {};
        for (const shareKey in this.reports) {
            const transferReports = [];
            const shareReport = this.reports[shareKey];
            for (const key in shareReport) {
                const transferReport = shareReport[key];
                transferReports.push(transferReport);
            }
            report[shareKey] = transferReports;
        }
        return report;
    }
    hasQueuedTransfer(hash, kind) {
        for (const waiting of this.waiting) {
            if (waiting.hash === hash && waiting.kind === kind) {
                return true;
            }
        }
        for (const active of this.active) {
            if (active.hash === hash && active.kind === kind) {
                return true;
            }
        }
        for (const complete of this.completed) {
            if (complete.hash === hash && complete.kind === kind) {
                return true;
            }
        }
        return false;
    }
    onReportUpdate(cb) {
        return this.reportBus.on(cb);
    }
    transfersRequestedByUsFinished() {
        return this.transfersRequestedByUsEnroller.isDone();
    }
}
