"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseAddress = exports.parseShareAddress = exports.parseAuthorAddress = exports.parseAuthorOrShareAddress = exports.checkShareIsValid = exports.checkAuthorIsValid = exports.assembleShareAddress = exports.assembleAuthorAddress = void 0;
const errors_js_1 = require("../util/errors.js");
const characters_js_1 = require("./characters.js");
//================================================================================
/** Put a short name and pub key together into an identity address. */
function assembleAuthorAddress(name, encodedPubkey) {
    return `@${name}.${encodedPubkey}`;
}
exports.assembleAuthorAddress = assembleAuthorAddress;
/** Put a share name and encoded pub key together into a share address. */
function assembleShareAddress(name, encodedPubkey) {
    return `+${name}.${encodedPubkey}`;
}
exports.assembleShareAddress = assembleShareAddress;
/** Check that an identity address is valid. */
function checkAuthorIsValid(addr) {
    const parsed = parseAuthorAddress(addr);
    if ((0, errors_js_1.notErr)(parsed))
        return true;
    return parsed;
}
exports.checkAuthorIsValid = checkAuthorIsValid;
/** Check that a share address is valid. */
function checkShareIsValid(addr) {
    const parsed = parseShareAddress(addr);
    if ((0, errors_js_1.notErr)(parsed))
        return true;
    return parsed;
}
exports.checkShareIsValid = checkShareIsValid;
/** Parse a share or author address into its parts */
function parseAuthorOrShareAddress(address) {
    if (address.startsWith("@")) {
        return parseAuthorAddress(address);
    }
    else if (address.startsWith("+")) {
        return parseShareAddress(address);
    }
    return new errors_js_1.ValidationError("address must start with either @ or +");
}
exports.parseAuthorOrShareAddress = parseAuthorOrShareAddress;
/** Parse an author address into its parts. */
function parseAuthorAddress(address) {
    return parseAddress(address, {
        sigil: "@",
        separator: ".",
        minNameLength: 4,
        maxNameLength: 4,
        minPubkeyLength: 53,
        maxPubkeyLength: 53,
        allowedNameCharacters: characters_js_1.authorNameChars,
        allowedPubkeyCharacters: characters_js_1.authorKeyChars,
        pubkeyMustStartWithB: true,
    });
}
exports.parseAuthorAddress = parseAuthorAddress;
/** Parse a share address into its parts. */
function parseShareAddress(address) {
    return parseAddress(address, {
        sigil: "+",
        separator: ".",
        minNameLength: 1,
        maxNameLength: 15,
        minPubkeyLength: 53,
        maxPubkeyLength: 53,
        allowedNameCharacters: characters_js_1.workspaceNameChars,
        allowedPubkeyCharacters: characters_js_1.workspaceKeyChars,
        pubkeyMustStartWithB: false,
    });
}
exports.parseShareAddress = parseShareAddress;
function parseAddress(address, opts) {
    const { sigil, separator, minNameLength, maxNameLength, minPubkeyLength, maxPubkeyLength, allowedNameCharacters, allowedPubkeyCharacters, pubkeyMustStartWithB, } = opts;
    if (typeof address !== "string") {
        return new errors_js_1.ValidationError("address must be a string");
    }
    if (address.length < 4)
        return new errors_js_1.ValidationError("address is too short");
    if (address[0] !== sigil) {
        return new errors_js_1.ValidationError(`address must start with a sigil: "${sigil}"`);
    }
    if (address.indexOf(separator) === -1) {
        return new errors_js_1.ValidationError(`address must contain a separator character: "${separator}"`);
    }
    const parts = address.slice(1).split(separator);
    if (parts.length !== 2) {
        return new errors_js_1.ValidationError(`address must have exactly 2 parts separated by a "${separator}" separator`);
    }
    const [name, pubkey] = parts;
    if (name.length < minNameLength || name.length > maxNameLength) {
        return new errors_js_1.ValidationError(`name must be between ${minNameLength} and ${maxNameLength} characters long, but is ${name.length}`);
    }
    if (pubkey.length < minPubkeyLength || pubkey.length > maxPubkeyLength) {
        return new errors_js_1.ValidationError(`pubkey must be between ${minPubkeyLength} and ${maxPubkeyLength} characters long, but is ${pubkey.length}`);
    }
    if (!(0, characters_js_1.onlyHasChars)(name, allowedNameCharacters)) {
        return new errors_js_1.ValidationError(`name "${name}" must only have allowed characters`);
    }
    if (!(0, characters_js_1.onlyHasChars)(pubkey, allowedPubkeyCharacters)) {
        return new errors_js_1.ValidationError(`pubkey "${pubkey}" must only have allowed characters`);
    }
    if ((0, characters_js_1.isDigit)(name[0])) {
        return new errors_js_1.ValidationError(`name "${name}" must not start with a digit`);
    }
    if ((0, characters_js_1.isDigit)(pubkey[0])) {
        return new errors_js_1.ValidationError(`pubkey "${pubkey}" must not start with a digit`);
    }
    if (pubkeyMustStartWithB && pubkey[0] !== "b") {
        return new errors_js_1.ValidationError(`pubkey "${pubkey}" must start with 'b'`);
    }
    return {
        address,
        name,
        pubkey,
    };
}
exports.parseAddress = parseAddress;
