"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.checkObj = exports.checkInt = exports.checkString = exports.checkLiteral = exports.checkIsPlainObject = exports.isPlainObject = void 0;
const bytes_js_1 = require("../util/bytes.js");
const characters_js_1 = require("./characters.js");
//================================================================================
function isPlainObject(obj) {
    if (Object.prototype.toString.call(obj) !== "[object Object]") {
        return false;
    }
    // reject class instances
    if (("" + obj.constructor).startsWith("class"))
        return false;
    return true;
}
exports.isPlainObject = isPlainObject;
function checkIsPlainObject(x) {
    return isPlainObject(x) ? null : "expected plain object but got " + x;
}
exports.checkIsPlainObject = checkIsPlainObject;
//================================================================================
function checkLiteral(val) {
    return (x) => {
        if (x !== val)
            return `expected literal value ${JSON.stringify(val)}`;
        return null;
    };
}
exports.checkLiteral = checkLiteral;
function checkString(opts = {}) {
    return ((x) => {
        if (opts.optional !== true && x === undefined)
            return "required";
        if (opts.optional === true && x === undefined)
            return null; // skip the rest of the checks if it's undefined
        //let len = x.length;  // TODO: string length or byte length?
        let len = (0, bytes_js_1.stringLengthInBytes)(x);
        if (typeof x !== "string") {
            return "expected a string but got " + JSON.stringify(x);
        }
        if (opts.minLen !== undefined && len < opts.minLen) {
            return `string shorter than min length of ${opts.minLen} chars`;
        }
        if (opts.maxLen !== undefined && len > opts.maxLen) {
            return `string shorter than max length of ${opts.maxLen} chars`;
        }
        if (opts.len !== undefined && len !== opts.len) {
            return `string does not have required length of ${opts.len} chars: ${x}`;
        }
        if (opts.allowedChars !== undefined &&
            !(0, characters_js_1.onlyHasChars)(x, opts.allowedChars)) {
            return "contains disallowed characters";
        }
        return null;
    });
}
exports.checkString = checkString;
function checkInt(opts = {}) {
    return ((x) => {
        if (opts.optional !== true && x === undefined)
            return "required";
        if (opts.optional === true && x === undefined)
            return null; // skip the rest of the checks if it's undefined
        if (opts.nullable !== true && x === null)
            return "not nullable";
        if (opts.nullable === true && x === null)
            return null; // skip the rest of the checks if it's null
        if (typeof x !== "number") {
            return "expected a number but got " + JSON.stringify(x);
        }
        if (isNaN(x))
            return "is NaN";
        if (!isFinite(x))
            return "is Infinity";
        if (x !== Math.round(x))
            return "expected an integer";
        if (opts.min !== undefined && x < opts.min) {
            return `integer too small (must be >= ${opts.min})`;
        }
        if (opts.max !== undefined && x > opts.max) {
            return `integer too large (must be <= ${opts.max})`;
        }
        return null;
    });
}
exports.checkInt = checkInt;
function checkObj(opts = {}) {
    // set defaults
    opts.allowLiteralUndefined = opts.allowLiteralUndefined ?? false;
    opts.allowExtraKeys = opts.allowExtraKeys ?? false;
    return (x) => {
        if (!isPlainObject(x))
            return "expected an object";
        if (opts.allowLiteralUndefined === false) {
            // don't allow any keys to be literally undefined, e.g. exiting but set to undefined.
            // when this is false, optional keys have to be omitted instead of set to undefined.
            for (let [k, v] of Object.entries(x)) {
                if (v === undefined) {
                    return `${k} is explicitly set to undefined but should be missing instead`;
                }
            }
        }
        if (opts.objSchema !== undefined) {
            // look for extra keys
            if (opts.allowExtraKeys === false) {
                let objKeys = Object.keys(x);
                let schemaKeys = Object.keys(opts.objSchema);
                let extraObjKeys = objKeys.filter((k) => schemaKeys.indexOf(k) === -1);
                if (extraObjKeys.length > 0) {
                    return `object has extra keys not in the schema: ${extraObjKeys.join(", ")}`;
                }
            }
            // check the individual schemas of each key
            for (let [key, validator] of Object.entries(opts.objSchema)) {
                let err = validator(x[key]);
                if (err !== null)
                    return `${key}: ${err}`;
            }
        }
        return null;
    };
}
exports.checkObj = checkObj;
