"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CryptoDriverChloride = void 0;
const chloride_1 = __importDefault(require("chloride"));
const buffer_1 = require("buffer");
const crypto_1 = __importDefault(require("crypto"));
const bytes_js_1 = require("../util/bytes.js");
const buffers_js_1 = require("../util/buffers.js");
//--------------------------------------------------
const log_js_1 = require("../util/log.js");
const updatable_hash_js_1 = require("./updatable_hash.js");
let logger = new log_js_1.Logger("crypto-driver-chloride", "cyan");
(0, log_js_1.setLogLevel)("crypto-driver-chloride", log_js_1.LogLevel.Info);
//================================================================================
/**
 * A version of the ICryptoDriver interface backed by Chloride.
 * Faster than noble.
 * Works in Node and the browser (with some polyfilling on your part).
 */
const CryptoDriverChloride = class {
    static sha256(input) {
        if (typeof input === "string")
            input = (0, buffers_js_1.stringToBuffer)(input);
        if ((0, bytes_js_1.identifyBufOrBytes)(input) === "bytes")
            input = (0, buffers_js_1.bytesToBuffer)(input);
        const resultBuf = chloride_1.default.crypto_hash_sha256(input);
        return Promise.resolve((0, buffers_js_1.bufferToBytes)(resultBuf));
    }
    static updatableSha256() {
        return new updatable_hash_js_1.UpdatableHash({
            hash: crypto_1.default.createHash("sha256"),
            update: (hash, data) => hash.update(data),
            digest: (hash) => {
                const digest = hash.digest();
                if (typeof digest === "string") {
                    return (0, bytes_js_1.stringToBytes)(digest);
                }
                return (0, buffers_js_1.bufferToBytes)(digest);
            },
        });
    }
    static generateKeypairBytes(seed) {
        // If provided, the seed is used as the secret key.
        // If omitted, a random secret key is generated.
        logger.debug("generateKeypairBytes");
        let seedBuf = seed === undefined ? undefined : (0, buffers_js_1.bytesToBuffer)(seed);
        if (!seedBuf) {
            seedBuf = buffer_1.Buffer.alloc(32);
            chloride_1.default.randombytes(seedBuf);
        }
        const keys = chloride_1.default.crypto_sign_seed_keypair(seedBuf);
        return Promise.resolve({
            //curve: 'ed25519',
            pubkey: (0, buffers_js_1.bufferToBytes)(keys.publicKey),
            // so that this works with either sodium or libsodium-wrappers (in browser):
            secret: (0, buffers_js_1.bufferToBytes)((keys.secretKey).slice(0, 32)),
        });
    }
    static sign(keypairBytes, msg) {
        logger.debug("sign");
        const secretBuf = (0, buffers_js_1.bytesToBuffer)((0, bytes_js_1.concatBytes)(keypairBytes.secret, keypairBytes.pubkey));
        if (typeof msg === "string")
            msg = (0, buffers_js_1.stringToBuffer)(msg);
        if (msg instanceof Uint8Array)
            msg = (0, buffers_js_1.bytesToBuffer)(msg);
        return Promise.resolve((0, buffers_js_1.bufferToBytes)(
        // this returns a Buffer
        chloride_1.default.crypto_sign_detached(msg, secretBuf)));
    }
    static verify(publicKey, sig, msg) {
        logger.debug("verify");
        try {
            if (typeof msg === "string")
                msg = (0, buffers_js_1.stringToBuffer)(msg);
            if (msg instanceof Uint8Array)
                msg = (0, buffers_js_1.bytesToBuffer)(msg);
            return Promise.resolve(chloride_1.default.crypto_sign_verify_detached((0, buffers_js_1.bytesToBuffer)(sig), msg, publicKey));
        }
        catch {
            /* istanbul ignore next */
            return Promise.resolve(false);
        }
    }
};
exports.CryptoDriverChloride = CryptoDriverChloride;
