"use strict";
// This file has no type annotations because we can't get the typings for Node's crypto library here.
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CryptoDriverNode = void 0;
const crypto_1 = __importDefault(require("crypto"));
const bytes_js_1 = require("../util/bytes.js");
const buffers_js_1 = require("../util/buffers.js");
//--------------------------------------------------
const log_js_1 = require("../util/log.js");
let logger = new log_js_1.Logger("crypto-driver-node", "cyan");
//================================================================================
const _generateKeypairDerBytes = () => {
    // Generate a keypair in "der" format, which we will have to process
    // to remove some prefixes.
    //
    // Typescript has outdated definitions, doesn't know about ed25519.
    // So fight it with "as any".
    let pair = crypto_1.default.generateKeyPairSync("ed25519", {
        publicKeyEncoding: {
            format: "der",
            type: "spki",
        },
        privateKeyEncoding: {
            format: "der",
            type: "pkcs8",
        },
    });
    // Typescript thinks these are strings, but they're Buffers...
    // and we need to convert them to bytes (uint8arrays)
    return {
        pubkey: (0, buffers_js_1.bufferToBytes)(pair.publicKey),
        secret: (0, buffers_js_1.bufferToBytes)(pair.privateKey),
    };
};
function _shortenDer(k) {
    return ({
        pubkey: k.pubkey.slice(-32),
        secret: k.secret.slice(-32),
    });
}
let _derPrefixPublic = (0, bytes_js_1.b64StringToBytes)("MCowBQYDK2VwAyEA");
let _derPrefixSecret = (0, bytes_js_1.b64StringToBytes)("MC4CAQAwBQYDK2VwBCIEIA==");
function _lengthenDerPublic(b) {
    return (0, bytes_js_1.concatBytes)(_derPrefixPublic, b);
}
function _lengthenDerSecret(b) {
    return (0, bytes_js_1.concatBytes)(_derPrefixSecret, b);
}
/**
 * A verison of the ICrptoDriver interface backed by native Node crypto functions.
 * Requires a recent version of Node, perhaps 12+?
 * Does not work in the browser.
 */
const CryptoDriverNode = class {
    static async sha256(input) {
        return (0, buffers_js_1.bufferToBytes)(crypto_1.default.createHash("sha256").update(input).digest());
    }
    static updatableSha256() {
        return new UpdatableHash({
            hash: crypto_1.default.createHash("sha256"),
            update: (hash, data) => hash.update(data),
            digest: (hash) => {
                const digest = hash.digest();
                if (typeof digest === "string") {
                    return stringToBytes(digest);
                }
                return (0, buffers_js_1.bufferToBytes)(digest);
            },
        });
    }
    static async generateKeypairBytes() {
        logger.debug("generateKeypairBytes");
        return _shortenDer(_generateKeypairDerBytes());
    }
    static async sign(keypairBytes, msg) {
        logger.debug("sign");
        if (typeof msg === "string")
            msg = (0, buffers_js_1.stringToBuffer)(msg);
        return (0, buffers_js_1.bufferToBytes)(crypto_1.default.sign(null, msg, {
            key: (0, buffers_js_1.bytesToBuffer)(_lengthenDerSecret(keypairBytes.secret)),
            format: "der",
            type: "pkcs8",
        }));
    }
    static async verify(publicKey, sig, msg) {
        logger.debug("verif");
        if (typeof msg === "string")
            msg = (0, buffers_js_1.stringToBuffer)(msg);
        try {
            return crypto_1.default.verify(null, msg, {
                key: _lengthenDerPublic(publicKey),
                format: "der",
                type: "spki",
            }, sig);
        }
        catch {
            return false;
        }
    }
};
exports.CryptoDriverNode = CryptoDriverNode;
