"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Crypto = void 0;
const errors_js_1 = require("../util/errors.js");
const misc_js_1 = require("../util/misc.js");
const base32_js_1 = require("./base32.js");
const keypair_js_1 = require("./keypair.js");
const addresses_js_1 = require("../core-validators/addresses.js");
const global_crypto_driver_js_1 = require("./global-crypto-driver.js");
//--------------------------------------------------
const log_js_1 = require("../util/log.js");
let logger = new log_js_1.Logger("crypto", "cyan");
//================================================================================
/** Higher-level crypto functions. Mostly used for generating new author and share keypairs. */
const Crypto = class {
    /** Do a sha256 hash, then return the output bytes encoded as base32. */
    static async sha256base32(input) {
        const b32 = await global_crypto_driver_js_1.GlobalCryptoDriver.sha256(input);
        return (0, base32_js_1.base32BytesToString)(b32);
    }
    static updatableSha256() {
        return global_crypto_driver_js_1.GlobalCryptoDriver.updatableSha256();
    }
    /**
     * Generate a new author keypair — a keypair of public and private keys as strings encoded in base32.
     *
     * NOTE: this will return a different keypair every time, even if the name is the same.
     * (Names are not unique.)
     *
     * Returns a ValidationError if the name doesn't follow the rules.
     *
     * @param shortname A 4-character nickname to make the address easier to remember and identify.
     */
    static async generateAuthorKeypair(shortname) {
        logger.debug(`generateAuthorKeypair("${shortname}")`);
        const keypairBytes = await global_crypto_driver_js_1.GlobalCryptoDriver
            .generateKeypairBytes();
        const keypairFormatted = {
            address: (0, addresses_js_1.assembleAuthorAddress)(shortname, (0, base32_js_1.base32BytesToString)(keypairBytes.pubkey)),
            secret: (0, base32_js_1.base32BytesToString)(keypairBytes.secret),
        };
        // Make sure it's valid (correct length, etc).  return error if invalid.
        const err = (0, addresses_js_1.checkAuthorIsValid)(keypairFormatted.address);
        if ((0, errors_js_1.isErr)(err))
            return err;
        return keypairFormatted;
    }
    /**
     * Generate a new share keypair — a keypair of public and private keys as strings encoded in base32.
     *
     * NOTE: this will return a different keypair every time, even if the name is the same.
     * (Names are not unique.)
     *
     * Returns a ValidationError if the name doesn't follow the rules.
     *
     * @param name A free-form name to identify the share.
     */
    static async generateShareKeypair(name) {
        logger.debug(`generateAuthorKeypair("${name}")`);
        const keypairBytes = await global_crypto_driver_js_1.GlobalCryptoDriver
            .generateKeypairBytes();
        const keypairFormatted = {
            shareAddress: (0, addresses_js_1.assembleShareAddress)(name, (0, base32_js_1.base32BytesToString)(keypairBytes.pubkey)),
            secret: (0, base32_js_1.base32BytesToString)(keypairBytes.secret),
        };
        // Make sure it's valid (correct length, etc).  return error if invalid.
        const err = (0, addresses_js_1.checkShareIsValid)(keypairFormatted.shareAddress);
        if ((0, errors_js_1.isErr)(err))
            return err;
        return keypairFormatted;
    }
    /**
     * Sign a message using an Earthstar keypair.
     * Return a signature as base32 string.
     *
     * Can return a ValidationError if the keypair is bad or something goes unexpectedly wrong with signing.
     */
    static async sign(keypair, msg) {
        logger.debug(`sign`);
        try {
            const keypairBytes = (0, keypair_js_1.decodeKeypairToBytes)(keypair);
            if ((0, errors_js_1.isErr)(keypairBytes))
                return keypairBytes;
            const signed = await global_crypto_driver_js_1.GlobalCryptoDriver.sign(keypairBytes, msg);
            return (0, base32_js_1.base32BytesToString)(signed);
        }
        catch (err) {
            /* istanbul ignore next */
            return new errors_js_1.ValidationError("unexpected error while signing: " + err.message);
        }
    }
    /**
     * Check if a ed25519 signature is valid.
     *
     * This returns false on any kind of failure:
     *   * bad author address format
     *   * bad signature base32 format
     *   * signature base32 format is valid but signature itself is invalid
     *   * unexpected failure from crypto library
     */
    static verify(address, sig, msg) {
        logger.debug(`verify`);
        try {
            const parsed = (0, addresses_js_1.parseAuthorOrShareAddress)(address);
            if ((0, errors_js_1.isErr)(parsed))
                return Promise.resolve(false);
            return global_crypto_driver_js_1.GlobalCryptoDriver.verify((0, base32_js_1.base32StringToBytes)(parsed.pubkey), (0, base32_js_1.base32StringToBytes)(sig), msg);
        }
        catch {
            // catch any unexpected errors
            return Promise.resolve(false);
        }
    }
    /**
     * Check if an ed25519 keypair is valid, e.g. does the secret match the pubkey.
     *
     * Returns...
     * - true on success (format is correct, and secret matches pubkey)
     * - a ValidationError if the secret does not match the pubkey.
     * - a ValidationError if the author address or secret are not validly formatted strings.
     * - a ValidationError if anything else goes wrong
     */
    static async checkKeypairIsValid(keypair) {
        // We check if the secret matches the pubkey by signing something and then validating the signature.
        // However, key generation is deterministic, so it would be more direct to just do this:
        //
        //     let pubkeyBytes = LowLevelCrypto.generateKeypairBytes(base32StringtoBytes(keypair.secret))
        //     then check if pubkeyBytes matches keypair.address
        //
        // ...but only some of the cryptodrivers let you give a seed for keypair generation.
        // ...so this signature trick will work for now.
        logger.debug(`checkAuthorKeypairIsValid`);
        const address = (0, keypair_js_1.isAuthorKeypair)(keypair)
            ? keypair.address
            : keypair.shareAddress;
        try {
            if (typeof address !== "string" ||
                typeof keypair.secret !== "string") {
                return new errors_js_1.ValidationError("address and secret must be strings");
            }
            const parseErr = (0, addresses_js_1.parseAuthorOrShareAddress)(address);
            if ((0, errors_js_1.isErr)(parseErr))
                return parseErr;
            const msg = "a test message to sign. " + (0, misc_js_1.randomId)();
            const sig = await this.sign(keypair, msg);
            if ((0, errors_js_1.isErr)(sig))
                return sig;
            const isValid = await this.verify(address, sig, msg);
            if (isValid === false) {
                return new errors_js_1.ValidationError("pubkey does not match secret");
            }
            return true;
        }
        catch (err) {
            return new errors_js_1.ValidationError("unexpected error in checkAuthorKeypairIsValid: " + err.message);
        }
    }
};
exports.Crypto = Crypto;
