"use strict";
/**
 * [Earthstar](https://earthstar-project.org) is a small and resilient distributed storage protocol designed with a strong focus on simplicity and versatility, with the social realities of peer-to-peer computing kept in mind.
 *
 * This is a reference implementation written in Typescript. You can use it to add Earthstar functionality to applications running on servers, browsers, the command line, or anywhere else JavaScript can be run.
 *
 * ### Example usage
 *
 * ```ts
 * import { Replica, ReplicaDriverMemory, Crypto, Peer } from "earthstar";
 *
 * const shareKeypair = await Crypto.generateShareKeypair("gardening");
 *
 * const replica = new Replica({
 * 	 driver: ReplicaDriverMemory(shareKeypair.shareAddress),
 * 	 shareSecret: shareKeypair.secret,
 * });
 *
 * const authorKeypair = await Crypto.generateAuthorKeypair("suzy");
 *
 * await replica.set(authorKeypair, {
 * 	 path: "/my-note",
 * 	 text: "Saw seven magpies today",
 * });
 *
 * const allDocs = await replica.getAllDocs();
 *
 * const peer = new Peer();
 *
 * peer.addReplica(replica);
 *
 * peer.sync("https://my.server")
 * ```
 *
 * This module also exposes server APIs for for building always-online peers. The below example reads some on-disk JSON to initiate some share replicas, and stores their data using the filesystem.
 *
 * ```ts
 * import {
 * ExtensionKnownShares,
 * ExtensionSyncWeb,
 * Server,
 * } from "https://deno.land/x/earthstar/mod.ts";
 *
 * const server = new Server([
 * new ExtensionKnownShares({
 * 	 knownSharesPath: "./known_shares.json",
 * 	 onCreateReplica: (shareAddress) => {
 * 		 return new Earthstar.Replica({
 * 			 driver: new ReplicaDriverFs(shareAddress, "./share_data"),
 * 		 });
 * 	 },
 * }),
 * new ExtensionSyncWebsocket(),
 * ]);
 *
 * @module
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __exportStar = (this && this.__exportStar) || function(m, exports) {
    for (var p in m) if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p)) __createBinding(exports, m, p);
};
Object.defineProperty(exports, "__esModule", { value: true });
__exportStar(require("../core-validators/addresses.js"), exports);
__exportStar(require("../core-validators/characters.js"), exports);
__exportStar(require("../core-validators/checkers.js"), exports);
__exportStar(require("../crypto/base32.js"), exports);
__exportStar(require("../crypto/crypto-driver-noble.js"), exports);
__exportStar(require("../crypto/crypto-types.js"), exports);
__exportStar(require("../crypto/crypto.js"), exports);
__exportStar(require("../crypto/global-crypto-driver.js"), exports);
__exportStar(require("../crypto/keypair.js"), exports);
__exportStar(require("../formats/format_es4.js"), exports);
__exportStar(require("../formats/format_es5.js"), exports);
__exportStar(require("../formats/util.js"), exports);
__exportStar(require("../formats/format_types.js"), exports);
__exportStar(require("../syncer/syncer.js"), exports);
__exportStar(require("../syncer/partner_local.js"), exports);
__exportStar(require("../syncer/partner_web_client.js"), exports);
__exportStar(require("../syncer/syncer_types.js"), exports);
__exportStar(require("../peer/peer-types.js"), exports);
__exportStar(require("../peer/peer.js"), exports);
__exportStar(require("../query/query-types.js"), exports);
__exportStar(require("../query/query.js"), exports);
__exportStar(require("../query/query-helpers.js"), exports);
__exportStar(require("../replica/compare.js"), exports);
__exportStar(require("../replica/replica.js"), exports);
__exportStar(require("../replica/replica_cache.js"), exports);
__exportStar(require("../replica/replica-types.js"), exports);
__exportStar(require("../replica/multiformat_replica.js"), exports);
__exportStar(require("../replica/util-types.js"), exports);
__exportStar(require("../replica/doc_drivers/memory.js"), exports);
__exportStar(require("../replica/attachment_drivers/memory.js"), exports);
__exportStar(require("../replica/driver_memory.js"), exports);
__exportStar(require("../util/bytes.js"), exports);
__exportStar(require("../util/doc-types.js"), exports);
__exportStar(require("../util/errors.js"), exports);
__exportStar(require("../util/invite.js"), exports);
__exportStar(require("../util/log.js"), exports);
__exportStar(require("../util/misc.js"), exports);
__exportStar(require("../util/shared_settings.js"), exports);
