"use strict";
var _a;
Object.defineProperty(exports, "__esModule", { value: true });
exports.FormatEs4 = void 0;
const errors_js_1 = require("../util/errors.js");
const crypto_js_1 = require("../crypto/crypto.js");
const characters_js_1 = require("../core-validators/characters.js");
const checkers_js_1 = require("../core-validators/checkers.js");
const addresses_js_1 = require("../core-validators/addresses.js");
//--------------------------------------------------
const log_js_1 = require("../util/log.js");
let logger = new log_js_1.Logger("validator es.4", "red");
// Tolerance for accepting messages from the future (because of clock skew between peers)
const FUTURE_CUTOFF_MINUTES = 10;
const FUTURE_CUTOFF_MICROSECONDS = FUTURE_CUTOFF_MINUTES * 60 * 1000 * 1000;
// Allowed valid range of timestamps (in microseconds, not milliseconds)
const MIN_TIMESTAMP = 10000000000000; // 10^13
const MAX_TIMESTAMP = 9007199254740990; // Number.MAX_SAFE_INTEGER - 1
const MAX_CONTENT_LENGTH = 4000000; // 4 million bytes = 4 megabytes (measured as bytes of utf-8, not normal string length)
const HASH_STR_LEN = 53; // number of base32 characters including leading 'b', which is 32 raw bytes when decoded
const SIG_STR_LEN = 104; // number of base32 characters including leading 'b', which is 64 raw bytes when decoded
const ES4_CORE_SCHEMA = {
    objSchema: {
        format: (0, checkers_js_1.checkLiteral)("es.4"),
        author: (0, checkers_js_1.checkString)({ allowedChars: characters_js_1.authorAddressChars }),
        content: (0, checkers_js_1.checkString)({ maxLen: MAX_CONTENT_LENGTH }),
        contentHash: (0, checkers_js_1.checkString)({ allowedChars: characters_js_1.b32chars, len: HASH_STR_LEN }),
        deleteAfter: (0, checkers_js_1.checkInt)({
            min: MIN_TIMESTAMP,
            max: MAX_TIMESTAMP,
            nullable: true,
        }),
        path: (0, checkers_js_1.checkString)({ allowedChars: characters_js_1.pathChars, minLen: 2, maxLen: 512 }),
        signature: (0, checkers_js_1.checkString)({ allowedChars: characters_js_1.b32chars, len: SIG_STR_LEN }),
        timestamp: (0, checkers_js_1.checkInt)({ min: MIN_TIMESTAMP, max: MAX_TIMESTAMP }),
        workspace: (0, checkers_js_1.checkString)({ allowedChars: characters_js_1.workspaceAddressChars }),
    },
    allowLiteralUndefined: false,
    allowExtraKeys: false,
};
/**
 * Validator for the 'es.4' format. Checks if documents are spec-compliant before ingesting, and signs them according to spec.
 * @link https://earthstar-project.org/specs/data-spec-es4
 */
exports.FormatEs4 = (_a = class {
        /** Deterministic hash of this version of the document */
        static hashDocument(doc) {
            // Deterministic hash of the document.
            // Can return a ValidationError, but only checks for very basic document validity.
            // The hash of the document is used for signatures and references to specific docs.
            // We use the hash of the content in case we want to drop the actual content
            // and only keep the hash around for verifying signatures.
            // None of these fields are allowed to contain tabs or newlines
            // (except content, but we use contentHash instead).
            // to check the basic validity it needs a signature of the correct length and characters,
            // but the actual content of the signature is not checked here.
            // so let's fake it.
            const docWithFakeSig = {
                ...doc,
                signature: "bthisisafakesignatureusedtofillintheobjectwhenvalidatingitforhashingaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            };
            const err = this._checkBasicDocumentValidity(docWithFakeSig);
            if ((0, errors_js_1.isErr)(err))
                return Promise.resolve(err);
            // Sort fields in lexicographic order by field name.
            // let result = ''
            // For each field,
            //     skip "content" and "signature" fields.
            //     skip fields with value === null.
            //     result += fieldname + "\t" + convertToString(value) + "\n"
            // return base32encode(sha256(result).binaryDigest())
            return crypto_js_1.Crypto.sha256base32(`author\t${doc.author}\n` +
                `contentHash\t${doc.contentHash}\n` +
                (doc.deleteAfter === null
                    ? ""
                    : `deleteAfter\t${doc.deleteAfter}\n`) +
                `format\t${doc.format}\n` +
                `path\t${doc.path}\n` +
                `timestamp\t${doc.timestamp}\n` +
                `workspace\t${doc.workspace}\n`);
        }
        /**
         * Generate a signed document from the input format the validator expects.
         */
        static async generateDocument({ input, keypair, share, timestamp }) {
            const doc = {
                format: "es.4",
                author: keypair.address,
                content: input.content,
                contentHash: await crypto_js_1.Crypto.sha256base32(input.content),
                deleteAfter: input.deleteAfter ?? null,
                path: input.path,
                timestamp,
                workspace: share,
                signature: "?", // signature will be added in just a moment
                // _localIndex will be added during upsert.  it's not needed for the signature.
            };
            const signedDoc = await this.signDocument(keypair, doc);
            if ((0, errors_js_1.isErr)(signedDoc)) {
                return signedDoc;
            }
            return { doc: signedDoc };
        }
        /**
         * Generate a signed document from the input format the validator expects.
         */
        static async signDocument(keypair, doc) {
            const hash = await this.hashDocument(doc);
            if ((0, errors_js_1.isErr)(hash))
                return hash;
            const sig = await crypto_js_1.Crypto.sign(keypair, hash);
            if ((0, errors_js_1.isErr)(sig))
                return sig;
            return { ...doc, signature: sig };
        }
        /**
         * Overwrite the user-written contents of a document, wipe any associated data, and return the signed document.
         */
        static async wipeDocument(keypair, doc) {
            if (doc.content.length === 0) {
                return doc;
            }
            const cleanedResult = this.removeExtraFields(doc);
            if ((0, errors_js_1.isErr)(cleanedResult))
                return cleanedResult;
            const cleanedDoc = cleanedResult.doc;
            // make new doc which is empty and just barely newer than the original
            const emptyDoc = {
                ...cleanedDoc,
                content: "",
                contentHash: await crypto_js_1.Crypto.sha256base32(""),
                signature: "?",
            };
            return this.signDocument(keypair, emptyDoc);
        }
        /**
         * Return a copy of the doc without extra fields, plus the extra fields
         * as a separate object.
         * If the input is not a plain javascript object, return a ValidationError.
         * This should be run before checkDocumentIsValid.  The output doc will be
         * more likely to be valid once the extra fields have been removed.
         */
        static removeExtraFields(doc) {
            if (!(0, checkers_js_1.isPlainObject)(doc)) {
                return new errors_js_1.ValidationError("doc is not a plain javascript object");
            }
            const validKeys = new Set(Object.keys(ES4_CORE_SCHEMA.objSchema || {}));
            const doc2 = {};
            const extras = {};
            for (const [key, val] of Object.entries(doc)) {
                if (validKeys.has(key)) {
                    doc2[key] = val;
                }
                else {
                    if (!key.startsWith("_")) {
                        return new errors_js_1.ValidationError("extra document fields must have names starting with an underscore");
                    }
                    extras[key] = val;
                }
            }
            return {
                doc: doc2,
                extras,
            };
        }
        /**
         * This calls all the more detailed functions which start with underscores.
         * Returns true if the document is ok, or returns a ValidationError if anything is wrong.
         * Normally `now` should be omitted so that it defaults to the current time,
         * or you can override it for testing purposes.
         */
        static async checkDocumentIsValid(doc, now) {
            if (now === undefined)
                now = Date.now() * 1000;
            // do this first to ensure we have all the right datatypes in the right fields
            const errBV = this._checkBasicDocumentValidity(doc);
            if ((0, errors_js_1.isErr)(errBV))
                return errBV;
            // this is the most likely to fail under regular conditions, so do it next
            // (because of clock skew and expired ephemeral documents)
            const errT = this._checkTimestampIsOk(doc.timestamp, doc.deleteAfter, now);
            if ((0, errors_js_1.isErr)(errT))
                return errT;
            const errW = this._checkAuthorCanWriteToPath(doc.author, doc.path);
            if ((0, errors_js_1.isErr)(errW))
                return errW;
            const errP = this._checkPathIsValid(doc.path, doc.deleteAfter);
            if ((0, errors_js_1.isErr)(errP))
                return errP;
            const errAA = (0, addresses_js_1.checkAuthorIsValid)(doc.author);
            if ((0, errors_js_1.isErr)(errAA))
                return errAA;
            const errWA = (0, addresses_js_1.checkShareIsValid)(doc.workspace);
            if ((0, errors_js_1.isErr)(errWA))
                return errWA;
            // do this after validating that the author address is well-formed
            // so we don't pass garbage into the crypto signature code
            const errS = await this._checkAuthorSignatureIsValid(doc);
            if ((0, errors_js_1.isErr)(errS))
                return errS;
            // do this last since it might be slow on a large document
            const errCH = await this._checkContentMatchesHash(doc.content, doc.contentHash);
            if ((0, errors_js_1.isErr)(errCH))
                return errCH;
            return true;
        }
        // These are broken out for easier unit testing.
        // They will not normally be used directly; use the main assertDocumentIsValid instead.
        // Return true on success.
        static _checkBasicDocumentValidity(doc) {
            const err = (0, checkers_js_1.checkObj)(ES4_CORE_SCHEMA)(doc);
            if (err !== null)
                return new errors_js_1.ValidationError(err);
            return true;
        }
        static _checkAuthorCanWriteToPath(author, path) {
            // Can the author write to the path?
            // return a ValidationError, or return true on success.
            // no tilde: it's public, anyone can write
            if (path.indexOf("~") === -1)
                return true;
            // path contains "~" + author.  the author can write here.
            if (path.indexOf("~" + author) !== -1)
                return true;
            // else, path contains at least one tilde but not ~@author.  The author can't write here.
            return new errors_js_1.ValidationError(`author ${author} can't write to path ${path}`);
        }
        static _checkTimestampIsOk(timestamp, deleteAfter, now) {
            // Check for valid timestamp, and expired ephemeral documents.
            // return a ValidationError, or return true on success.
            // timestamp and deleteAfter are already verified as good numbers by the schema checker:
            // - in the right range of min and max allowed timestamps
            // - integers, and not NaN or infinity
            // Timestamp must not be from the future.
            if (timestamp > now + FUTURE_CUTOFF_MICROSECONDS) {
                return new errors_js_1.ValidationError("timestamp too far in the future");
            }
            // Ephemeral documents
            if (deleteAfter !== null) {
                // Only valid if expiration date is in the future
                if (now > deleteAfter) {
                    return new errors_js_1.ValidationError("ephemeral doc has expired");
                }
                // Can't expire before it was created, that makes no sense
                if (deleteAfter <= timestamp) {
                    return new errors_js_1.ValidationError("ephemeral doc expired before it was created");
                }
            }
            return true;
        }
        static _checkPathIsValid(path, deleteAfter) {
            // Ensure the path matches the spec for allowed path strings.
            //
            // Path validity depends on if the document is ephemeral or not.  To check
            // that rule, supply deleteAfter.  Omit deleteAfter to skip checking that rule
            // (e.g. to just check if a path is potentially valid, ephemeral or not).
            //
            // return a ValidationError, or return true on success.
            // A path is a series of one or more path segments.
            // A path segment is '/' followed by one or more allowed characters.
            // the schema already checked that this
            // - is a string
            // - length between 2 and 512 characters inclusive
            // - onlyHasChars(pathChars)
            if (!path.startsWith("/")) {
                return new errors_js_1.ValidationError("invalid path: must start with /");
            }
            if (path.endsWith("/")) {
                return new errors_js_1.ValidationError("invalid path: must not end with /");
            }
            if (path.startsWith("/@")) {
                // This is disallowed so that we can tell paths and authors apart
                // when joining a workspace and a path/author in a URL:
                // +gardening.xxxxx/@aaaa.xxxx
                // +gardening.xxxxx/wiki/shared/Bumblebee
                return new errors_js_1.ValidationError('invalid path: must not start with "/@"');
            }
            if (path.indexOf("//") !== -1) {
                return new errors_js_1.ValidationError("invalid path: must not contain two consecutive slashes");
            }
            if (deleteAfter !== undefined) {
                // path must contain at least one '!', if and only if the document is ephemeral
                if (path.indexOf("!") === -1 && deleteAfter !== null) {
                    return new errors_js_1.ValidationError("when deleteAfter is set, path must contain '!'");
                }
                if (path.indexOf("!") !== -1 && deleteAfter === null) {
                    return new errors_js_1.ValidationError("when deleteAfter is null, path must not contain '!'");
                }
            }
            return true;
        }
        static async _checkAuthorSignatureIsValid(doc) {
            // Check if the signature is good.
            // return a ValidationError, or return true on success.
            try {
                const hash = await this.hashDocument(doc);
                if ((0, errors_js_1.isErr)(hash))
                    return hash;
                const verified = await crypto_js_1.Crypto.verify(doc.author, doc.signature, hash);
                if (verified !== true) {
                    return new errors_js_1.ValidationError("signature is invalid");
                }
                return true;
            }
            catch {
                return new errors_js_1.ValidationError("signature is invalid (unexpected exception)");
            }
        }
        static async _checkContentMatchesHash(content, contentHash) {
            // Ensure the contentHash matches the actual content.
            // return a ValidationError, or return true on success.
            // TODO: if content is null, skip this check
            if (await crypto_js_1.Crypto.sha256base32(content) !== contentHash) {
                return new errors_js_1.ValidationError("content does not match contentHash");
            }
            return true;
        }
        static getAttachmentInfo(_doc) {
            return new errors_js_1.ValidationError("es.4 does not support attachments");
        }
        static updateAttachmentFields(_creds, _doc, _size, _hash) {
            return Promise.resolve(new errors_js_1.ValidationError("es.4 does not support attachments"));
        }
        static authorFromCredentials(credentials) {
            return credentials.address;
        }
    },
    Object.defineProperty(_a, "id", {
        enumerable: true,
        configurable: true,
        writable: true,
        value: "es.4"
    }),
    _a);
