"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Peer = void 0;
//--------------------------------------------------
const log_js_1 = require("../util/log.js");
const stream_utils_js_1 = require("../streams/stream_utils.js");
const partner_web_server_js_1 = require("../syncer/partner_web_server.js");
const partner_local_js_1 = require("../syncer/partner_local.js");
const syncer_manager_js_1 = require("../syncer/syncer_manager.js");
const logger = new log_js_1.Logger("peer", "orangeRed");
const J = JSON.stringify;
//================================================================================
/** Holds many shares' replicas and manages their synchronisation with other peers. Recommended as the point of contact between your application and Earthstar shares.
 *
 * ```ts
 * const peer = new Peer();
 * peer.addReplica(myReplica);
 * peer.sync("https://my.server");
 * ```
 */
class Peer {
    constructor() {
        Object.defineProperty(this, "replicaEventBus", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new stream_utils_js_1.BlockingBus()
        });
        Object.defineProperty(this, "syncerManager", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /** A map of the replicas stored in this peer. */
        Object.defineProperty(this, "replicaMap", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Map()
        });
        logger.debug("constructor");
        this.syncerManager = new syncer_manager_js_1.SyncerManager(this);
    }
    //--------------------------------------------------
    // getters
    hasShare(share) {
        return this.replicaMap.has(share);
    }
    /** Returns an array of all the addresses of this peer's held replicas. */
    shares() {
        const keys = [...this.replicaMap.keys()];
        keys.sort();
        return keys;
    }
    replicas() {
        const keys = [...this.replicaMap.keys()];
        keys.sort();
        return keys.map((key) => this.replicaMap.get(key));
    }
    /** The number of replicas held by this peer */
    size() {
        return this.replicaMap.size;
    }
    getReplica(ws) {
        return this.replicaMap.get(ws);
    }
    //--------------------------------------------------
    // setters
    async addReplica(replica) {
        logger.debug(`addReplica(${J(replica.share)})`);
        if (this.replicaMap.has(replica.share)) {
            logger.debug(`already had a replica with that share`);
            throw new Error(`Peer.addReplica: already has a replica with share ${J(replica.share)}.  Don't add another one.`);
        }
        this.replicaMap.set(replica.share, replica);
        await this.replicaEventBus.send(this.replicaMap);
        logger.debug(`    ...addReplica: done`);
    }
    /** Remove a replica from the peer using its share address. */
    async removeReplicaByShare(share) {
        logger.debug(`removeReplicaByShare(${J(share)})`);
        this.replicaMap.delete(share);
        await this.replicaEventBus.send(this.replicaMap);
    }
    async removeReplica(replica) {
        const existingReplica = this.replicaMap.get(replica.share);
        if (replica === existingReplica) {
            logger.debug(`removeReplica(${J(replica.share)})`);
            await this.removeReplicaByShare(replica.share);
        }
        else {
            logger.debug(`removeReplica(${J(replica.share)}) -- same share but it's a different instance now; ignoring`);
        }
    }
    //----------------------------------------------
    // Syncing stuff
    /**
     * Begin synchronising with a remote or local peer.
     * @param target - A HTTP URL or `Peer` instance.
     * @param continuous - Whether the connection should be kept open for new changes from the other peer, or stop after initial reconciliation. Defaults to false.
     * @param formats - Optional. Which document formats to sync. Defaults to `es.5`.
     */
    sync(target, continuous = false, formats) {
        try {
            const partner = new partner_web_server_js_1.PartnerWebServer({
                url: target,
                appetite: continuous ? "continuous" : "once",
            });
            return this.syncerManager.addPartner(partner, target, formats);
        }
        catch {
            if (target instanceof Peer) {
                const partner = new partner_local_js_1.PartnerLocal(target, this, continuous ? "continuous" : "once", formats);
                return this.syncerManager.addPartner(partner, "Local", formats);
            }
            // This shouldn't happen.
            console.error("Provided an invalid target for syncing to a peer:", target);
            return undefined;
        }
    }
    /** Begin syncing using an instance implementing `ISyncPartner`. Use this if you don't want to sync with a local peer or a server. */
    addSyncPartner(partner, description, formats) {
        return this.syncerManager.addPartner(partner, description, formats);
    }
    /** Get all syncers created by this peer. This includes running and cancelled syncers. */
    getSyncers() {
        return this.syncerManager.getSyncers();
    }
    //----------------------------------------------
    // Subscribe stuff
    /** Fires a given callback whenever the Peer's store of replicas changes. */
    onReplicasChange(callback) {
        return this.replicaEventBus.on(callback);
    }
    /** Fires a given callback whenever a new syncer is added to the peer. */
    onSyncersChange(callback) {
        return this.syncerManager.onSyncersChange(callback);
    }
}
exports.Peer = Peer;
