"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.MultiformatReplica = void 0;
const dntShim = __importStar(require("../../_dnt.shims.js"));
const util_types_js_1 = require("./util-types.js");
const errors_js_1 = require("../util/errors.js");
const misc_js_1 = require("../util/misc.js");
const compare_js_1 = require("./compare.js");
const addresses_js_1 = require("../core-validators/addresses.js");
//--------------------------------------------------
const log_js_1 = require("../util/log.js");
const stream_utils_js_1 = require("../streams/stream_utils.js");
const util_js_1 = require("../formats/util.js");
const query_js_1 = require("../query/query.js");
const deps_js_1 = require("../../deps.js");
const J = JSON.stringify;
const logger = new log_js_1.Logger("replica", "gold");
const loggerSet = new log_js_1.Logger("replica set", "gold");
const loggerIngest = new log_js_1.Logger("replica ingest", "gold");
//================================================================================
function docCompareNewestFirst(a, b) {
    // Sorts by timestamp DESC (newest fist) and breaks ties using the signature ASC.
    return (0, compare_js_1.compareArrays)([a.timestamp, a.signature], [b.timestamp, a.signature], ["DESC", "ASC"]);
}
/**
 * A replica holding a share's data, used to read, write, and synchronise data to.
 *
 * Should be closed using the `close` method when no longer being used.
 *
 * ```ts
 * const gardeningKeypair = await Crypto.generateShareKeypair("gardening");
 *
 * const myReplica = new Replica({
 *  driver: new ReplicaDriverMemory(gardeningKeypair.shareAddress),
 *  config: {
 *    "es.5": { shareSecret: gardeningKeypair.secret }
 *  }
 * });
 * ```
 */
class MultiformatReplica {
    constructor({ driver, config }) {
        Object.defineProperty(this, "replicaId", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        }); // todo: save it to the driver too, and reload it when starting up
        /** The address of the share this replica belongs to. */
        Object.defineProperty(this, "share", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        /** The validator used to validate ingested documents. */
        Object.defineProperty(this, "replicaDriver", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "formatsConfig", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_isClosed", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: false
        });
        Object.defineProperty(this, "ingestLockStream", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new stream_utils_js_1.LockStream()
        });
        Object.defineProperty(this, "eventMultiStream", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new stream_utils_js_1.ChannelMultiStream("kind", true)
        });
        Object.defineProperty(this, "eventWriter", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "callbackSink", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new stream_utils_js_1.CallbackSink()
        });
        Object.defineProperty(this, "eraseInterval", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "expireEventTimeouts", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Map()
        });
        const addressIsValidResult = (0, addresses_js_1.checkShareIsValid)(driver.docDriver.share);
        if ((0, errors_js_1.isErr)(addressIsValidResult)) {
            throw addressIsValidResult;
        }
        logger.debug(`constructor.  driver = ${driver?.constructor?.name}`);
        this.replicaId = "replica-" + (0, misc_js_1.randomId)();
        this.share = driver.docDriver.share;
        this.replicaDriver = driver;
        this.formatsConfig = config || {};
        this.eventWriter = this.eventMultiStream.getWritableStream().getWriter();
        this.eventMultiStream.getReadableStream("*").pipeTo(new dntShim.WritableStream(this.callbackSink));
        this.eraseInterval = dntShim.setInterval(() => {
            if (!this.isClosed()) {
                this.pruneExpiredDocsAndAttachments();
            }
            else {
                clearInterval(this.eraseInterval);
            }
        }, 1000 * 60 * 60);
    }
    //--------------------------------------------------
    // LIFECYCLE
    /** Returns whether the replica is closed or not. */
    isClosed() {
        return this._isClosed;
    }
    /**
     * Closes the replica, preventing new documents from being ingested or events being emitted.
     * Any methods called after closing will return `ReplicaIsClosedError`.
     * @param erase - Erase the contents of the replica. Defaults to `false`.
     */
    async close(erase) {
        logger.debug("closing...");
        if (this._isClosed)
            throw new errors_js_1.ReplicaIsClosedError();
        // TODO: do this all in a lock?
        // Make sure writing lock is clear
        await this.ingestLockStream.close();
        for (const timeout of this.expireEventTimeouts.values()) {
            clearTimeout(timeout);
        }
        logger.debug("    sending willClose blockingly...");
        await this.eventWriter.write({
            kind: "willClose",
        });
        logger.debug("    marking self as closed...");
        if (erase === false) {
            await this.pruneExpiredDocsAndAttachments();
        }
        await this.replicaDriver.attachmentDriver.clearStaging();
        this._isClosed = true;
        logger.debug(`    closing ReplicaDriver (erase = ${erase})...`);
        await this.replicaDriver.docDriver.close(erase);
        await this.replicaDriver.attachmentDriver.close(erase);
        clearInterval(this.eraseInterval);
        logger.debug("    sending didClose nonblockingly...");
        await this.eventWriter.write({
            kind: "didClose",
        });
        logger.debug("...closing done");
        return Promise.resolve();
    }
    //--------------------------------------------------
    // CONFIG
    async getConfig(key) {
        if (this._isClosed)
            throw new errors_js_1.ReplicaIsClosedError();
        return await this.replicaDriver.docDriver.getConfig(key);
    }
    async setConfig(key, value) {
        if (this._isClosed)
            throw new errors_js_1.ReplicaIsClosedError();
        return await this.replicaDriver.docDriver.setConfig(key, value);
    }
    async listConfigKeys() {
        if (this._isClosed)
            throw new errors_js_1.ReplicaIsClosedError();
        return await this.replicaDriver.docDriver.listConfigKeys();
    }
    async deleteConfig(key) {
        if (this._isClosed)
            throw new errors_js_1.ReplicaIsClosedError();
        return await this.replicaDriver.docDriver.deleteConfig(key);
    }
    //--------------------------------------------------
    // GET
    /** Returns the max local index of all stored documents */
    getMaxLocalIndex() {
        if (this._isClosed)
            throw new errors_js_1.ReplicaIsClosedError();
        return this.replicaDriver.docDriver.getMaxLocalIndex();
    }
    /** Returns all documents, including historical versions of documents by other identities. */
    getAllDocs(formats) {
        logger.debug(`getAllDocs()`);
        return this.queryDocs({
            historyMode: "all",
            orderBy: "path ASC",
        }, formats);
    }
    /** Returns latest document from every path. */
    getLatestDocs(formats) {
        logger.debug(`getLatestDocs()`);
        return this.queryDocs({
            historyMode: "latest",
            orderBy: "path ASC",
        }, formats);
    }
    /** Returns all versions of a document by different authors from a specific path. */
    getAllDocsAtPath(path, formats) {
        logger.debug(`getAllDocsAtPath("${path}")`);
        return this.queryDocs({
            historyMode: "all",
            orderBy: "path ASC",
            filter: { path: path },
        }, formats);
    }
    /** Returns the most recently written version of a document at a path. */
    async getLatestDocAtPath(path, format) {
        logger.debug(`getLatestDocsAtPath("${path}")`);
        const docs = await this.queryDocs({
            historyMode: "latest",
            orderBy: "path ASC",
            filter: { path: path },
        }, format ? [format] : undefined);
        if (docs.length === 0)
            return undefined;
        return docs[0];
    }
    /** Returns an array of docs for a given query.
      ```
      const myQuery = {
        filter: {
          pathEndsWith: ".txt"
        },
        limit: 5,
      };
  
      const firstFiveTextDocs = await myReplica.queryDocs(myQuery);
      ```
      */
    async queryDocs(query = {}, formats) {
        logger.debug(`queryDocs`, query);
        if (this._isClosed)
            throw new errors_js_1.ReplicaIsClosedError();
        const f = (0, util_js_1.getFormatsWithFallback)(formats);
        return await this.replicaDriver.docDriver.queryDocs({
            ...query,
            formats: f.map((f) => f.id),
        });
    }
    /** Returns an array of all unique paths of documents returned by a given query. */
    async queryPaths(query = {}, formats) {
        const docs = await this.queryDocs(query, formats);
        const pathsSet = new Set(docs.map(({ path }) => path));
        return Array.from(pathsSet).sort();
    }
    /** Returns an array of all unique authors of documents returned by a given query. */
    async queryAuthors(query = {}, formats) {
        const docs = await this.queryDocs(query, formats);
        const authorsSet = new Set(docs.map(({ author }) => author));
        return Array.from(authorsSet).sort();
    }
    //--------------------------------------------------
    // SET
    /**
     * Adds a new document to the replica. If a document signed by the same identity exists at the same path, it will be overwritten.
     */
    // The Input type should match the formatter.
    // The default format should be es5
    async set(keypair, docToSet, format = util_js_1.DEFAULT_FORMAT) {
        loggerSet.debug(`set`, docToSet);
        if (this._isClosed)
            throw new errors_js_1.ReplicaIsClosedError();
        loggerSet.debug("...deciding timestamp: getting latest doc at the same path (from any author)");
        const latestDocSamePath = await this.getLatestDocAtPath(docToSet.path);
        let timestamp;
        if (typeof docToSet.timestamp === "number") {
            timestamp = docToSet.timestamp;
        }
        else {
            // bump timestamp if needed to win over existing latest doc at same path
            if (latestDocSamePath === undefined) {
                timestamp = (0, misc_js_1.microsecondNow)();
            }
            else {
                timestamp = Math.max((0, misc_js_1.microsecondNow)(), latestDocSamePath.timestamp + 1);
            }
        }
        loggerSet.debug("...generating doc");
        let cleanedDoc;
        if (latestDocSamePath) {
            const res = format.removeExtraFields(latestDocSamePath);
            if (!(0, errors_js_1.isErr)(res)) {
                cleanedDoc = res.doc;
            }
        }
        const result = await format.generateDocument({
            keypair,
            input: { ...docToSet, format: format.id },
            share: this.share,
            timestamp,
            prevLatestDoc: cleanedDoc,
            config: this.formatsConfig[format.id],
        });
        if ((0, errors_js_1.isErr)(result)) {
            return {
                kind: "failure",
                reason: "invalid_document",
                err: result,
            };
        }
        loggerSet.debug("...signature =", result.doc.signature);
        // The result has provided a new attachment for us to ingest.
        // The lack of this does not indicate that no attachment is associated with this doc
        // (it may refer to the attachment from the previous doc)
        if (result.attachment) {
            // Stage the new attachment with the attachment driver.
            const stageResult = await this.replicaDriver.attachmentDriver.stage(format.id, result.attachment);
            if ((0, errors_js_1.isErr)(stageResult)) {
                return {
                    kind: "failure",
                    reason: `invalid_document`,
                    err: stageResult,
                };
            }
            // Update the document's attachment fields using the results derived from staging.
            const updatedDocRes = await format.updateAttachmentFields(keypair, result.doc, stageResult.size, stageResult.hash, this.formatsConfig[format.id]);
            if ((0, errors_js_1.isErr)(updatedDocRes)) {
                await stageResult.reject();
                return {
                    kind: "failure",
                    reason: `invalid_document`,
                    err: updatedDocRes,
                };
            }
            // If everything checks out, commit the staged attachment to storage.
            loggerSet.debug("...ingesting attachment");
            loggerSet.debug("-----------------------");
            await stageResult.commit();
            await this.eventWriter.write({
                kind: "attachment_ingest",
                doc: updatedDocRes,
                hash: stageResult.hash,
                size: stageResult.size,
                sourceId: "local",
            });
            loggerSet.debug("...done ingesting attachment");
            loggerSet.debug("...ingesting");
            loggerSet.debug("-----------------------");
            // And ingest the document.
            const ingestEvent = await this.ingest(format, updatedDocRes, "local");
            loggerSet.debug("...done ingesting");
            loggerSet.debug("...set is done.");
            return ingestEvent;
        }
        // We don't need to do anything with attachments, so just ingest the document.
        loggerSet.debug("...ingesting");
        loggerSet.debug("-----------------------");
        const ingestResult = await this.ingest(format, result.doc, "local");
        loggerSet.debug("...done ingesting");
        loggerSet.debug("...set is done.");
        return ingestResult;
    }
    /**
     * Ingest an existing signed document to the replica.
     */
    async ingest(format, docToIngest, sourceId) {
        loggerIngest.debug(`ingest`, docToIngest);
        if (this._isClosed)
            throw new errors_js_1.ReplicaIsClosedError();
        loggerIngest.debug("...removing extra fields");
        const removeResultsOrErr = format
            .removeExtraFields(docToIngest);
        if ((0, errors_js_1.isErr)(removeResultsOrErr)) {
            return {
                kind: "failure",
                reason: "invalid_document",
                err: removeResultsOrErr,
            };
        }
        docToIngest = removeResultsOrErr.doc; // a copy of doc without extra fields
        const extraFields = removeResultsOrErr.extras; // any extra fields starting with underscores
        if (Object.keys(extraFields).length > 0) {
            loggerIngest.debug(`...extra fields found: ${J(extraFields)}`);
        }
        const docIsValid = await format.checkDocumentIsValid(docToIngest);
        if ((0, errors_js_1.isErr)(docIsValid)) {
            return {
                kind: "failure",
                reason: "invalid_document",
                err: docIsValid,
            };
        }
        const ingestPromise = (0, deps_js_1.deferred)();
        await this.ingestLockStream.run(async () => {
            // get other docs at the same path
            loggerIngest.debug(" >> ingest: start of protected region");
            loggerIngest.debug("  > getting other history docs at the same path by any author");
            const existingDocsSamePath = await this.getAllDocsAtPath(docToIngest.path, [format]);
            loggerIngest.debug(`  > ...got ${existingDocsSamePath.length}`);
            loggerIngest.debug("  > getting prevLatest and prevSameAuthor");
            const prevLatest = existingDocsSamePath[0] ?? null;
            const prevSameAuthor = existingDocsSamePath.filter((d) => d.author === docToIngest.author)[0] ??
                null;
            loggerIngest.debug("  > checking if new doc is latest at this path");
            existingDocsSamePath.push(docToIngest);
            existingDocsSamePath.sort(docCompareNewestFirst);
            const isLatest = existingDocsSamePath[0] === docToIngest;
            loggerIngest.debug(`  > ...isLatest: ${isLatest}`);
            if (!isLatest && prevSameAuthor !== null) {
                loggerIngest.debug("  > new doc is not latest and there is another one from the same author...");
                // check if this is obsolete or redudant from the same author
                const docComp = docCompareNewestFirst(docToIngest, prevSameAuthor);
                if (docComp === util_types_js_1.Cmp.GT) {
                    loggerIngest.debug("  > new doc is GT prevSameAuthor, so it is obsolete");
                    ingestPromise.resolve({
                        kind: "nothing_happened",
                        reason: "obsolete_from_same_author",
                        doc: docToIngest,
                    });
                    return;
                }
                if (docComp === util_types_js_1.Cmp.EQ) {
                    loggerIngest.debug("  > new doc is EQ prevSameAuthor, so it is redundant (already_had_it)");
                    ingestPromise.resolve({
                        kind: "nothing_happened",
                        reason: "already_had_it",
                        doc: docToIngest,
                    });
                    return;
                }
            }
            // save it
            loggerIngest.debug("  > upserting into ReplicaDriver...");
            const docAsWritten = await this.replicaDriver.docDriver.upsert(docToIngest);
            // TODO: pass existingDocsSamePath to save another lookup
            loggerIngest.debug("  > ...done upserting into ReplicaDriver");
            loggerIngest.debug("  > ...getting ReplicaDriver maxLocalIndex...");
            const maxLocalIndex = await this.replicaDriver.docDriver
                .getMaxLocalIndex();
            loggerIngest.debug(" >> ingest: end of protected region, returning a WriteEvent from the lock");
            const event = {
                kind: "success",
                maxLocalIndex,
                doc: docAsWritten,
                docIsLatest: isLatest,
                prevDocFromSameAuthor: prevSameAuthor,
                prevLatestDoc: prevLatest,
                sourceId,
            };
            ingestPromise.resolve(event);
            await this.eventWriter.write(event);
            if (docAsWritten.deleteAfter) {
                // Check for any previous delete after
                const key = `${docAsWritten.path}|${docAsWritten.author}`;
                const existingTimeout = this.expireEventTimeouts.get(key);
                if (existingTimeout) {
                    clearTimeout(existingTimeout);
                }
                const waitFor = (docAsWritten.deleteAfter / 1000) - Date.now();
                this.expireEventTimeouts.set(key, dntShim.setTimeout(() => {
                    this.eventWriter.write({
                        kind: "expire",
                        doc: docAsWritten,
                    });
                }, waitFor));
            }
        });
        return ingestPromise;
    }
    /**
     * Overwrite every document from this author, including history versions, with an empty doc.
      @returns The number of documents changed, or -1 if there was an error.
     */
    async overwriteAllDocsByAuthor(keypair, format) {
        const f = (0, util_js_1.getFormatWithFallback)(format);
        logger.debug(`overwriteAllDocsByAuthor("${keypair.address}")`);
        if (this._isClosed)
            throw new errors_js_1.ReplicaIsClosedError();
        // TODO: stream the docs out, overwrite them.
        const docsToOverwrite = await this.queryDocs({
            filter: { author: keypair.address },
            historyMode: "all",
        }, [f]);
        logger.debug(`    ...found ${docsToOverwrite.length} docs to overwrite`);
        let numOverwritten = 0;
        let numAlreadyEmpty = 0;
        for (const doc of docsToOverwrite) {
            const didWipe = await this.wipeDocument(keypair, doc, (0, util_js_1.getFormatWithFallback)(format));
            if ((0, errors_js_1.isErr)(didWipe)) {
                return didWipe;
            }
            else {
                // success
                numOverwritten += 1;
            }
        }
        logger.debug(`    ...done; ${numOverwritten} overwritten to be empty; ${numAlreadyEmpty} were already empty; out of total ${docsToOverwrite.length} docs`);
        return numOverwritten;
    }
    /** Wipe all content from a document at a given path, and erase its attachment (if it has one). */
    async wipeDocAtPath(keypair, path, format = util_js_1.DEFAULT_FORMAT) {
        const latestDocAtPath = await this.getLatestDocAtPath(path, format);
        if (!latestDocAtPath) {
            return new errors_js_1.ValidationError("No document exists at that path");
        }
        return this.wipeDocument(keypair, latestDocAtPath, format);
    }
    async wipeDocument(keypair, doc, format) {
        const docToWipe = {
            ...doc,
            timestamp: Math.max(doc.timestamp + 1, Date.now() * 1000),
            author: keypair.address,
        };
        const wipedDoc = await format.wipeDocument(keypair, docToWipe, this.formatsConfig[format.id]);
        if ((0, errors_js_1.isErr)(wipedDoc)) {
            return {
                kind: "failure",
                err: wipedDoc,
                reason: "invalid_document",
            };
        }
        const didIngest = await this.ingest(format, wipedDoc, "local");
        if (didIngest.kind === "success") {
            // Check if this document has an attachment by tring to get attachment info.
            const attachmentInfo = format.getAttachmentInfo(doc);
            if (!(0, errors_js_1.isErr)(attachmentInfo)) {
                // Wipe the attachment.
                // Ignore error indicating no attachment was found.
                const eraseRes = await this.replicaDriver.attachmentDriver.erase(format.id, attachmentInfo.hash);
                if (!(0, errors_js_1.isErr)(eraseRes)) {
                    await this.eventWriter.write({
                        kind: "attachment_prune",
                        format: format.id,
                        hash: attachmentInfo.hash,
                    });
                }
            }
        }
        return didIngest;
    }
    /** Erases expired docs and dangling attachments */
    async pruneExpiredDocsAndAttachments(
    // Because es5 is the only format with attachments, that's all we'll handle for now.
    formats = util_js_1.DEFAULT_FORMATS) {
        // Erase expired docs
        const erasedDocs = await this.replicaDriver.docDriver.eraseExpiredDocs();
        for (const doc of erasedDocs) {
            await this.eventWriter.write({
                kind: "expire",
                doc,
            });
        }
        // Erase dangling docs
        const formatLookup = (0, util_js_1.getFormatLookup)(formats);
        const allowedHashes = {};
        await this.getQueryStream({
            historyMode: "all",
            orderBy: "localIndex ASC",
        }, "existing", formats).pipeTo(new dntShim.WritableStream({
            write(event) {
                if (event.kind === "existing") {
                    const format = formatLookup[event.doc.format];
                    const attachmentInfo = format.getAttachmentInfo(event.doc);
                    if (!(0, errors_js_1.isErr)(attachmentInfo)) {
                        const maybeExistingSet = allowedHashes[format.id];
                        if (maybeExistingSet) {
                            maybeExistingSet.add(attachmentInfo.hash);
                        }
                        else {
                            allowedHashes[format.id] = new Set([attachmentInfo.hash]);
                        }
                    }
                }
            },
        }));
        const erasedAttachments = await this.replicaDriver.attachmentDriver.filter(allowedHashes);
        for (const attachment of erasedAttachments) {
            await this.eventWriter.write({
                kind: "attachment_prune",
                format: attachment.format,
                hash: attachment.hash,
            });
        }
    }
    /**
     * Returns a readable stream of replica events, such as new ingestions, document expirations, or the replica preparing to close.
     * @param channel - An optional string representing a channel of events to be subscribed to. Defaults to return all events.
     */
    getEventStream(channel = "*") {
        return this.eventMultiStream.getReadableStream(channel);
    }
    /**
     * Returns a readable stream of document events which match a given query. The events can represent existing documents, newly ingested documents, or expiring documents.
     */
    getQueryStream(query = {}, mode, formats) {
        const queryDocs = this.queryDocs.bind(this);
        const getEventStream = this.getEventStream.bind(this);
        return new dntShim.ReadableStream({
            async start(controller) {
                if (mode === "existing" || mode === "everything") {
                    const docs = await queryDocs(query, formats);
                    for (const doc of docs) {
                        controller.enqueue({
                            kind: "existing",
                            doc: doc,
                        });
                    }
                }
                controller.enqueue({ kind: "processed_all_existing" });
                if (mode === "existing") {
                    controller.close();
                    return;
                }
                const eventStream = getEventStream();
                const reader = eventStream.getReader();
                while (true) {
                    const { done, value: event } = await reader.read();
                    if (done)
                        return;
                    if (event.kind === "expire" || event.kind === "success") {
                        if (query.filter) {
                            if ((0, query_js_1.docMatchesFilter)(event.doc, query.filter)) {
                                controller.enqueue(event);
                                continue;
                            }
                        }
                        controller.enqueue(event);
                        continue;
                    }
                    if (event.kind === "didClose") {
                        controller.close();
                    }
                }
            },
        });
    }
    /**
     * Provide a callback to be triggered every time a replica event occurs.
     * @returns A callback which unsubscribes the event.
     */
    onEvent(callback) {
        return this.callbackSink.onWrite(callback);
    }
    //--------------------------------------------------
    // ATTACHMENTS
    /** Ingest an attachment for a given document.
     *
     * @returns `true` (indicating it was upsert), `false` (indicating this attachment is already in storage), or a `ValidationError` (indicating something went wrong.)
     */
    async ingestAttachment(format, doc, attachment, sourceId) {
        if (this._isClosed)
            throw new errors_js_1.ReplicaIsClosedError();
        const removeResultsOrErr = format
            .removeExtraFields(doc);
        if ((0, errors_js_1.isErr)(removeResultsOrErr)) {
            return Promise.resolve(removeResultsOrErr);
        }
        doc = removeResultsOrErr.doc; // a copy of doc without extra fields
        // check doc is valid
        const docIsValid = await format.checkDocumentIsValid(doc);
        if ((0, errors_js_1.isErr)(docIsValid)) {
            return Promise.resolve(docIsValid);
        }
        // Check we don't already have this attachment
        const existingAttachment = await this.getAttachment(doc, format);
        if (existingAttachment && !(0, errors_js_1.isErr)(existingAttachment)) {
            return false;
        }
        const attachmentInfo = format.getAttachmentInfo(doc);
        if ((0, errors_js_1.isErr)(attachmentInfo)) {
            // This really shouldn't happen, but...
            return Promise.resolve(attachmentInfo);
        }
        const stageRes = await this.replicaDriver.attachmentDriver
            .stage(doc.format, attachment);
        if ((0, errors_js_1.isErr)(stageRes)) {
            return stageRes;
        }
        // Compare the staged attachment's hash and size to what the doc claims it should be.
        if (stageRes.hash !== attachmentInfo.hash) {
            await stageRes.reject();
            return new errors_js_1.ValidationError("Attachment's hash did not match the document's");
        }
        if (stageRes.size !== attachmentInfo.size) {
            await stageRes.reject();
            return new errors_js_1.ValidationError("Attachment's size did not match the document's");
        }
        // If it all checks out, commit.
        await stageRes.commit();
        await this.eventWriter.write({
            kind: "attachment_ingest",
            doc,
            hash: stageRes.hash,
            size: stageRes.size,
            sourceId,
        });
        return true;
    }
    /** Gets an attachment for a given document. Returns a `ValidationError` if the given document can't have an attachment.
     */
    getAttachment(doc, format = util_js_1.DEFAULT_FORMAT) {
        if (doc.deleteAfter && doc.deleteAfter < Date.now() * 1000) {
            return Promise.resolve(new errors_js_1.ValidationError("This document has expired"));
        }
        const attachmentInfo = format.getAttachmentInfo(doc);
        if (!(0, errors_js_1.isErr)(attachmentInfo)) {
            return this.replicaDriver.attachmentDriver.getAttachment(doc.format, attachmentInfo.hash);
        }
        else {
            return Promise.resolve(attachmentInfo);
        }
    }
    /** Returns the given array of documents with a new `attachment` property merged in. The value of this property can be:
     * - `DocAttachment`
     * - `undefined` (the associated document can have an attachment, but we don't have a copy)
     * - `ValidationError` (the associated document can't have an attachment)
     */
    addAttachments(docs, formats) {
        const f = (0, util_js_1.getFormatsWithFallback)(formats);
        const formatLookup = {};
        for (const format of f) {
            formatLookup[format.id] = format;
        }
        const promises = docs.map((doc) => {
            return new Promise((resolve) => {
                const format = formatLookup[doc.format];
                const attachmentInfo = format.getAttachmentInfo(doc);
                if (!(0, errors_js_1.isErr)(attachmentInfo)) {
                    this.replicaDriver.attachmentDriver.getAttachment(doc.format, attachmentInfo.hash)
                        .then((attachment) => {
                        resolve({ ...doc, attachment });
                    });
                }
                else {
                    return resolve({
                        ...doc,
                        attachment: attachmentInfo,
                    });
                }
            });
        });
        return Promise.all(promises);
    }
}
exports.MultiformatReplica = MultiformatReplica;
