"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PartnerLocal = void 0;
const deps_js_1 = require("../../deps.js");
const errors_js_1 = require("../util/errors.js");
const errors_js_2 = require("../util/errors.js");
const syncer_js_1 = require("./syncer.js");
/** A syncing partner to be used with local instances of `IPeer`.
 * Works everywhere.
 */
class PartnerLocal {
    /**
     * @param peer - The target peer to sync with.
     * @param peerSelf - Our own peer.
     */
    constructor(peer, peerSelf, appetite, formats) {
        Object.defineProperty(this, "concurrentTransfers", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: 1024
        });
        Object.defineProperty(this, "payloadThreshold", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: 1
        });
        Object.defineProperty(this, "rangeDivision", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: 2
        });
        Object.defineProperty(this, "syncAppetite", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "outgoingQueue", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new deps_js_1.AsyncQueue()
        });
        Object.defineProperty(this, "incomingQueue", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new deps_js_1.AsyncQueue()
        });
        Object.defineProperty(this, "partnerPeer", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        // Need this for testing.
        Object.defineProperty(this, "partnerSyncer", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        this.syncAppetite = appetite;
        this.partnerPeer = peer;
        const { incomingQueue, outgoingQueue } = this;
        // This is a bit confusing, but it does work.
        // This will be the partner of our partner, which is us.
        // Now we create another syncer within this driver
        // But this syncer needs its own driver...
        // We'll give it one that proxies to the readable / writable pair we defined above.
        this.partnerSyncer = new syncer_js_1.Syncer({
            manager: peer.syncerManager,
            formats,
            partner: {
                syncAppetite: appetite,
                getEvents() {
                    return incomingQueue;
                },
                sendEvent(event) {
                    outgoingQueue.push(event);
                    return Promise.resolve();
                },
                concurrentTransfers: 1024,
                payloadThreshold: 32,
                rangeDivision: 32,
                async getDownload(opts) {
                    const partnerReplica = peerSelf.getReplica(opts.shareAddress);
                    if (!partnerReplica) {
                        throw new errors_js_2.ValidationError("Tried to get a receiving transfer for an unknown share.");
                    }
                    const attachment = await partnerReplica.replicaDriver.attachmentDriver
                        .getAttachment(opts.doc.format, opts.attachmentHash);
                    if (!attachment) {
                        return undefined;
                    }
                    if ((0, errors_js_1.isErr)(attachment)) {
                        return;
                    }
                    return await attachment.stream();
                },
                closeConnection() {
                    outgoingQueue.close();
                    return Promise.resolve();
                },
                handleUploadRequest(_opts) {
                    return Promise.resolve(new errors_js_1.NotSupportedError("PartnerLocal does not support uploads."));
                },
                handleTransferRequest(_source, _kind) {
                    // Don't need to implement this either.
                    return Promise.resolve(new errors_js_1.NotSupportedError("PartnerLocal does not support transfer requests."));
                },
            },
        });
    }
    getEvents() {
        return this.outgoingQueue;
    }
    sendEvent(event) {
        this.incomingQueue.push(event);
        return Promise.resolve();
    }
    closeConnection() {
        this.incomingQueue.close();
        return Promise.resolve();
    }
    async getDownload(opts) {
        const partnerReplica = this.partnerPeer.getReplica(opts.shareAddress);
        if (!partnerReplica) {
            throw new errors_js_2.ValidationError("Tried to get a receiving transfer for an unknown share.");
        }
        const attachment = await partnerReplica.replicaDriver.attachmentDriver
            .getAttachment(opts.doc.format, opts.attachmentHash);
        if (!attachment) {
            return undefined;
        }
        if ((0, errors_js_1.isErr)(attachment)) {
            return;
        }
        return await attachment.stream();
    }
    handleUploadRequest(_opts) {
        // Just return undefined here because we know how to directly get a transfer from this partner.
        return Promise.resolve(new errors_js_1.NotSupportedError("PartnerLocal does not support uploads."));
    }
    handleTransferRequest(_source, _kind) {
        // Don't need to implement this either.
        return Promise.resolve(new errors_js_1.NotSupportedError("PartnerLocal does not support transfer requests."));
    }
}
exports.PartnerLocal = PartnerLocal;
