"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PartnerWebClient = void 0;
const deps_js_1 = require("../../deps.js");
const stream_utils_js_1 = require("../streams/stream_utils.js");
const errors_js_1 = require("../util/errors.js");
/** A syncing partner created from an inbound HTTP connection (i.e. a web client).
 *
 * Works everywhere, but is really meant for servers running on Deno and Node.
 */
class PartnerWebClient {
    constructor({ socket, appetite }) {
        Object.defineProperty(this, "concurrentTransfers", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: 16
        });
        Object.defineProperty(this, "payloadThreshold", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: 8
        });
        Object.defineProperty(this, "rangeDivision", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: 8
        });
        Object.defineProperty(this, "syncAppetite", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "socket", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "incomingQueue", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new deps_js_1.AsyncQueue()
        });
        Object.defineProperty(this, "socketIsReady", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: (0, deps_js_1.deferred)()
        });
        this.syncAppetite = appetite;
        if (socket.readyState === socket.OPEN) {
            this.socketIsReady.resolve();
        }
        socket.onopen = () => {
            this.socketIsReady.resolve();
        };
        this.socket = socket;
        this.socket.binaryType = "arraybuffer";
        this.socket.onmessage = (event) => {
            // Casting as string for Node's incorrect WebSocket types.
            this.incomingQueue.push(JSON.parse(event.data));
        };
        this.socket.onclose = () => {
            this.incomingQueue.close();
        };
        this.socket.onerror = (event) => {
            if ("error" in event) {
                this.incomingQueue.close({
                    withError: event.error,
                });
                return;
            }
            this.incomingQueue.close({
                withError: new errors_js_1.EarthstarError("Websocket error."),
            });
        };
    }
    async sendEvent(event) {
        await this.socketIsReady;
        if (this.socket.readyState !== this.socket.OPEN) {
            return;
        }
        return this.socket.send(JSON.stringify(event));
    }
    getEvents() {
        return this.incomingQueue;
    }
    closeConnection() {
        this.socket.close();
        return Promise.resolve();
    }
    getDownload(_opts) {
        // Server can't initiate a request with a client.
        return Promise.resolve(new errors_js_1.NotSupportedError("SyncDriverWebServer does not support download requests."));
    }
    handleUploadRequest(_opts) {
        // Server won't get in-band BLOB_REQ messages
        return Promise.resolve(new errors_js_1.NotSupportedError("SyncDriverWebServer does not support upload requests."));
    }
    handleTransferRequest(socket, kind) {
        // Return a stream which writes to the socket. nice.
        //  They want to download data from us
        if (kind === "download") {
            const writable = (0, stream_utils_js_1.websocketWritable)(socket, (outgoing) => outgoing);
            return Promise.resolve(writable);
        }
        else {
            // they want to upload data to us.
            const readable = (0, stream_utils_js_1.websocketReadable)(socket, (event) => {
                if (event.data instanceof ArrayBuffer) {
                    const bytes = new Uint8Array(event.data);
                    return bytes;
                }
                return null;
            });
            return Promise.resolve(readable);
        }
    }
}
exports.PartnerWebClient = PartnerWebClient;
