"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TransferManager = void 0;
const deps_js_1 = require("../../deps.js");
const util_js_1 = require("../formats/util.js");
const stream_utils_js_1 = require("../streams/stream_utils.js");
const errors_js_1 = require("../util/errors.js");
const attachment_transfer_js_1 = require("./attachment_transfer.js");
const promise_enroller_js_1 = require("./promise_enroller.js");
const transfer_queue_js_1 = require("./transfer_queue.js");
class TransferManager {
    constructor(opts) {
        Object.defineProperty(this, "partner", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "queue", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "formats", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "otherSyncerId", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: (0, deps_js_1.deferred)()
        });
        Object.defineProperty(this, "receivedAllExpectedTransfersEnroller", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new promise_enroller_js_1.PromiseEnroller()
        });
        /** A kind of promise which resolves when all attachment requests have been made. */
        Object.defineProperty(this, "madeAllAttachmentRequestsEnroller", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new promise_enroller_js_1.PromiseEnroller(true)
        });
        Object.defineProperty(this, "formatsLookup", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "reportDidUpdateBus", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new stream_utils_js_1.BlockingBus()
        });
        Object.defineProperty(this, "expectedTransferPromises", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new Map()
        });
        this.partner = opts.partner;
        this.queue = new transfer_queue_js_1.TransferQueue(this.partner.concurrentTransfers);
        this.formats = opts.formats;
        this.formatsLookup = (0, util_js_1.getFormatLookup)(this.formats);
        this.queue.onReportUpdate(async (report) => {
            await this.reportDidUpdateBus.send(report);
        });
        this.madeAllAttachmentRequestsEnroller.isDone().then(() => {
            this.receivedAllExpectedTransfersEnroller.seal();
        });
        this.receivedAllExpectedTransfersEnroller.isDone().then(() => {
            this.queue.gotAllTransfersRequestedByUs();
        });
    }
    // pass a syncagent's isDone to this
    registerSyncAgent(agent) {
        this.madeAllAttachmentRequestsEnroller.enrol(agent.isDone());
    }
    // call this when all the sync agents have been constructed
    // and seal the enroller with all the piped existing docs.
    allSyncAgentsKnown() {
        this.madeAllAttachmentRequestsEnroller.seal();
    }
    registerExpectedTransfer(share, hash) {
        const key = `${share}_${hash}`;
        if (this.expectedTransferPromises.has(key)) {
            // We're already expecting this transfer, no need to add another promise.
            return;
        }
        const promise = (0, deps_js_1.deferred)();
        this.expectedTransferPromises.set(key, promise);
        this.receivedAllExpectedTransfersEnroller.enrol(promise);
    }
    async queueTransfer(transfer) {
        // Check if we already queued it from the queue
        if (this.queue.hasQueuedTransfer(transfer.hash, transfer.kind)) {
            return;
        }
        // Queue it up!
        await this.queue.addTransfer(transfer);
    }
    // This will be called by the sync agent.
    async handleDownload(doc, replica, counterpartId) {
        const format = this.formatsLookup[doc.format];
        const attachmentInfo = format.getAttachmentInfo(doc);
        if ((0, errors_js_1.isErr)(attachmentInfo)) {
            throw new errors_js_1.EarthstarError("TransferManager: attempted to download doc with no attachment.");
        }
        const result = await this.partner.getDownload({
            doc,
            shareAddress: replica.share,
            syncerId: await this.otherSyncerId,
            attachmentHash: attachmentInfo.hash,
        });
        // The partner doesn't have it.
        if (result === undefined) {
            /* I don't think this is needed...
            const key = `${replica.share}_${attachmentInfo.hash}`;
            const promise = deferred<void>();
            this.expectedTransferPromises.set(key, promise);
      
            promise.resolve();
            */
            return "no_attachment";
        }
        if ((0, errors_js_1.isErr)(result)) {
            return result;
        }
        const transfer = new attachment_transfer_js_1.AttachmentTransfer({
            replica,
            doc,
            format,
            stream: result,
            requester: "us",
            counterpartId,
        });
        await this.queueTransfer(transfer);
        return "queued";
    }
    async handleUpload(transferOpts, replica) {
        const format = this.formatsLookup[transferOpts.doc.format];
        const attachment = await replica.getAttachment(transferOpts.doc, format);
        if (!attachment) {
            return false;
        }
        const result = await this.partner.handleUploadRequest({
            shareAddress: transferOpts.shareAddress,
            syncerId: await this.otherSyncerId,
            doc: transferOpts.doc,
            attachmentHash: transferOpts.attachmentHash,
        });
        if ((0, errors_js_1.isErr)(result)) {
            return false;
        }
        try {
            const transfer = new attachment_transfer_js_1.AttachmentTransfer({
                doc: transferOpts.doc,
                format,
                replica,
                stream: result,
                requester: "them",
                counterpartId: "unused", // Doesn't matter
            });
            await this.queueTransfer(transfer);
        }
        catch {
            // This can happen if we don't have the attachment, or the doc can't have an attachment.
            return false;
        }
        return true;
    }
    async handleTransferRequest({ replica, path, author, source, kind, formatName, counterpartId }) {
        const stream = await this.partner.handleTransferRequest(source, kind);
        // This method is not supported or we just don't have the requested attachment.
        if (stream === undefined || (0, errors_js_1.isErr)(stream)) {
            return;
        }
        const format = (0, util_js_1.getFormatLookup)(this.formats)[formatName];
        if (!format) {
            return;
        }
        const docs = await replica.getAllDocsAtPath(path, [format]);
        const doc = docs.find((doc) => doc.author === author);
        if (!doc) {
            return;
        }
        // got a stream! add it to the syncer's transfers.
        const transfer = new attachment_transfer_js_1.AttachmentTransfer({
            doc: doc,
            format,
            replica,
            stream,
            requester: kind === "upload" ? "us" : "them",
            counterpartId,
        });
        if (transfer.requester === "us") {
            const key = `${transfer.share}_${transfer.hash}`;
            const promise = this.expectedTransferPromises.get(key);
            if (promise) {
                promise.resolve();
            }
        }
        await this.queueTransfer(transfer);
    }
    cancel() {
        this.queue.cancel();
    }
    getReport() {
        return this.queue.getReport();
    }
    onReportUpdate(cb) {
        return this.reportDidUpdateBus.on(cb);
    }
    transfersRequestedByUsFinished() {
        return this.queue.transfersRequestedByUsFinished();
    }
    registerOtherSyncerId(id) {
        this.otherSyncerId.resolve(id);
    }
    isAlreadyQueued(hash, kind) {
        return this.queue.hasQueuedTransfer(hash, kind);
    }
}
exports.TransferManager = TransferManager;
