"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseInvitationURL = exports.createInvitationURL = void 0;
const addresses_js_1 = require("../core-validators/addresses.js");
const crypto_js_1 = require("../crypto/crypto.js");
const errors_js_1 = require("./errors.js");
/** Creates an invitation URL. Validates the share address, that servers are valid URLs, and the secret against the share address if given. */
async function createInvitationURL(shareAddress, servers, secret) {
    const parsedShareAddress = (0, addresses_js_1.parseShareAddress)(shareAddress);
    if ((0, errors_js_1.isErr)(parsedShareAddress)) {
        return new errors_js_1.ValidationError(`Invalid share address.`);
    }
    for (const server of servers) {
        try {
            new URL(server);
        }
        catch {
            return new errors_js_1.ValidationError(`Could not parse ${server} as a URL.`);
        }
    }
    const serverParams = servers.map((url) => `&server=${url}`).join("&");
    let secretParam = "";
    if (secret) {
        const isValid = await crypto_js_1.Crypto.checkKeypairIsValid({
            shareAddress,
            secret,
        });
        if ((0, errors_js_1.isErr)(isValid)) {
            return new errors_js_1.ValidationError(`Supplied the wrong secret for ${shareAddress}`);
        }
        secretParam = `&secret=${secret}`;
    }
    const invitationURL = `earthstar://${shareAddress}/?invite${serverParams}${secretParam}&v=2`;
    return invitationURL;
}
exports.createInvitationURL = createInvitationURL;
/** Parses an invitation URL. Validates the share address, secret (if given), and any server URLs. */
async function parseInvitationURL(invitationURL) {
    try {
        const url = new URL(invitationURL);
        const isValidShareAddress = (0, addresses_js_1.parseShareAddress)(url.hostname);
        if ((0, errors_js_1.isErr)(isValidShareAddress)) {
            return new errors_js_1.ValidationError("Invitation did not include a valid share address.");
        }
        const params = new URLSearchParams(url.search);
        const isInvitation = params.get("invite");
        const version = params.get("v");
        if (isInvitation === null) {
            return new errors_js_1.ValidationError("Not an invitation URL");
        }
        if (version === null) {
            return new errors_js_1.ValidationError("Invitation version not specified.");
        }
        if (version !== "2") {
            return new errors_js_1.ValidationError(`Invitation version is ${version}, expected version 2.`);
        }
        const servers = params.getAll("server");
        for (const server of servers) {
            try {
                new URL(server);
            }
            catch {
                return new errors_js_1.ValidationError(`Invitation's servers included a malformed URL: ${server}`);
            }
        }
        const secret = params.get("secret");
        if (secret) {
            const isValid = await crypto_js_1.Crypto.checkKeypairIsValid({
                shareAddress: url.hostname,
                secret: secret,
            });
            if ((0, errors_js_1.isErr)(isValid)) {
                return new errors_js_1.ValidationError(`Invitation contains the wrong secret for share ${url.hostname}.`);
            }
        }
        return {
            shareAddress: url.hostname,
            secret: secret || undefined,
            servers: servers,
        };
    }
    catch {
        return new errors_js_1.ValidationError("Could not parse the invitation URL.");
    }
}
exports.parseInvitationURL = parseInvitationURL;
