"use strict";
/**
 * Logs are assigned a priority number.
 * Higher numbers are less important information.
 * Set your desired log level higher to get more info.
 *
 *  -1 nothing
 *   0 only errors
 *   1 also warnings
 *   2 also logs
 *   3 also debugs
 *
 * Logs also come from different "sources" which can
 * have different log level settings.
 *
 * Two ways to modify the log level settings:
 *
 * 1. Set an environment variable in the shell.
 *    This applies to all "sources" (can't be set individually)
 *         EARTHSTAR_LOG_LEVEL=2 npm run test
 *
 * 2. Use setLogLevels() to globally modify the levels:
 *         setLogLevels({ sync: 2 });
 *
 * The environment variable wins over the numbers set by setLogLevels.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.Logger = exports.getLogLevels = exports.getLogLevel = exports.setDefaultLogLevel = exports.setLogLevel = exports.updateLogLevels = exports.DEFAULT_LOG_LEVEL = exports.LogLevel = void 0;
var LogLevel;
(function (LogLevel) {
    LogLevel[LogLevel["None"] = -1] = "None";
    LogLevel[LogLevel["Error"] = 0] = "Error";
    LogLevel[LogLevel["Warn"] = 1] = "Warn";
    LogLevel[LogLevel["Log"] = 2] = "Log";
    LogLevel[LogLevel["Info"] = 3] = "Info";
    LogLevel[LogLevel["Debug"] = 4] = "Debug";
})(LogLevel = exports.LogLevel || (exports.LogLevel = {}));
exports.DEFAULT_LOG_LEVEL = LogLevel.Error;
//================================================================================
// ENV VAR
/*
// get the single log level number from the environment, or undefined if not set
const readEnvLogLevel = (): LogLevel | undefined => {
    if (process?.env?.EARTHSTAR_LOG_LEVEL) {
        const parsed = parseInt(process.env.EARTHSTAR_LOG_LEVEL);
        if (isNaN(parsed)) { return undefined }
        if (parsed !== Math.floor(parsed)) { return undefined; }
        return parsed;
    }
    return undefined;
}

// apply env var setting
const ENV_LOG_LEVEL = readEnvLogLevel();
*/
//================================================================================
// GLOBAL SETTINGS
// make global singleton to hold log levels
let globalLogLevels = {
    // result is the min of (_env) and anything else
    _default: exports.DEFAULT_LOG_LEVEL,
};
function updateLogLevels(newLogLevels) {
    globalLogLevels = {
        ...globalLogLevels,
        ...newLogLevels,
    };
}
exports.updateLogLevels = updateLogLevels;
function setLogLevel(source, level) {
    globalLogLevels[source] = level;
}
exports.setLogLevel = setLogLevel;
function setDefaultLogLevel(level) {
    globalLogLevels._default = level;
}
exports.setDefaultLogLevel = setDefaultLogLevel;
function getLogLevel(source) {
    if (source in globalLogLevels) {
        return globalLogLevels[source];
    }
    else {
        return globalLogLevels._default;
    }
}
exports.getLogLevel = getLogLevel;
function getLogLevels() {
    return globalLogLevels;
}
exports.getLogLevels = getLogLevels;
class Logger {
    constructor(source, color) {
        Object.defineProperty(this, "source", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "color", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: undefined
        });
        this.source = source;
        this.color = color || "aqua";
    }
    _print(level, showTag, indent, ...args) {
        if (level <= getLogLevel(this.source)) {
            if (showTag) {
                const tag = `[${this.source}]`;
                if (this.color !== undefined) {
                    const tagArgs = [`%c ${tag}`, `color: ${this.color}`];
                    console.log(indent, ...tagArgs, ...args);
                }
                else {
                    console.log(indent, tag, ...args);
                }
            }
            else {
                console.log(indent, ...args);
            }
        }
    }
    error(...args) {
        this._print(LogLevel.Error, true, "!!", ...args);
    }
    warn(...args) {
        this._print(LogLevel.Warn, true, "! ", ...args);
    }
    log(...args) {
        this._print(LogLevel.Log, true, "  ", ...args);
    }
    info(...args) {
        this._print(LogLevel.Info, true, "    ", ...args);
    }
    debug(...args) {
        this._print(LogLevel.Debug, true, "      ", ...args);
    }
    blank() {
        this._print(LogLevel.Info, false, "");
    }
}
exports.Logger = Logger;
