import { Query } from "./query-types.js";
import { Replica } from "../replica/replica.js";
import { DefaultFormats, FormatDocType, FormatsArg } from "../formats/format_types.js";
export declare function escapeStringForRegex(s: string): string;
export declare let _matchAll: (re: RegExp, str: string) => RegExpExecArray[];
/** A helper used by globToQueryAndRegex -- see that function for details.
 *
 *  This function simply turns a glob string into a regex.
 *  The other function calls this one, but sometimes discards the regex
 *   if it's not needed because it can make a good enough Earthstar query.
 */
export declare let globToRegex: (glob: string, forceEntireMatch?: boolean) => string;
/** Helper for querying Earthstar docs using a glob-style query string.
 *
 * Given a glob string, return:
 *    - an earthstar Query
 *    - and a regular expression (as a plain string, not a RegExp instance).
 *
 * Glob strings support '*' and '**' as wildcards.
 * '**' matches any sequence of characters at all including slashes.
 * '*' matches any sequence of characters except a forward slash --
 *  it does not span directories in the path.
 * Your glob string may have multiple asterisks in various positions,
 *  except they cannot be directly adjecent to each other (no '***' should
 *  ever occur -- this will cause a ValidationError to be thrown).
 *
 * Note that no other wildcards are supported, unlike Bash globs.
 *
 * To use this function, run the query yourself and apply the regex
 * as a filter to the paths of the resulting documents,
 *  to get only the documents whose paths match the glob.
 * The regex will be null if it's not needed (if the query is
 *  strong enough to get the job done by itself).
 *
 * The returned query will use some subset of
 *  the `path`, `pathStartsWith`, and `pathEndswith` properties,
 *  and no other properties.
 *
 * Example glob strings:
 *
 *     "/posts/*ing/*.json" matches "/posts/sailing/12345.json"
 *
 *     "/posts/**.json"     matches "/posts/sailing/12345.json"
 *                              and "/posts/a/b/c/d/e/f/g/12345.json"
 *                              and "/posts/.json"
 *
 *     "**"                 matches every possible path
 *
 * To use it:
 *
 *    let queryByGlob = async (replica: IReplica, glob: string): Promise<Doc[]> => {
 *       let { query, regex } = globToQueryAndRegex(glob);
 *
 *       let docs = await replica.queryDocs(query);
 *       if (regex != null) {
 *           let re = new RegExp(regex);
 *           docs = docs.filter(doc => re.test(doc.path));
 *       }
 *       return docs;
 *    }
 *
 *    let posts = await queryByGlob(myReplica, '/posts/*.txt');
 */
export declare let globToQueryAndRegex: (glob: string) => {
    query: Query<string[]>;
    regex: string | null;
};
/** Find documents whose path matches the glob string.
 * See documentation for globToQueryAndRegex for details on glob strings.
 *
 * You can specify additional query options by providing a `moreQueryOptions` object.
 * For example, you might want to set { contentLengthGt: 0 } to skip documents
 * with empty content (e.g. "deleted" documents).
 *
 * `moreQueryOptions` will override the glob's query, so it's best to avoid setting
 * `path`, `pathStartsWith` or `pathEndsWith` in your moreQueryOptions unless you
 * intend to override the glob's query.
 */
export declare function queryByGlob<F = DefaultFormats>(replica: Replica, glob: string, moreQueryOptions?: Omit<Query<[string]>, "formats">, formats?: FormatsArg<F>): Promise<FormatDocType<F>[]>;
interface ParsedTemplate {
    template: string;
    varNames: string[];
    glob: string;
    namedCaptureRegex: string;
}
/** This is a low-level helper for the template matching code; probably don't use it directly.
 *
 *  Given a template, parse it and return:
 *  - a list of variable names
 *  - a glob for searching Earthstar using queryByGlob()
 *  - a regular expression with named capture groups which can extract the
 *     values of the variables from a path.
 *
 *  A variable in the template is any alphanumeric chars or underscores, in curly braces
 *   like {example} or {__EXAMPLE__} or {example_1}, starting with a non-number.
 *
 *  Templates can also contain * and ** according to the glob rules.
 *   Those wildcards are not counted as variables but are allowed to expand during the
 *   regex phase.
 *
 *  Rules for templates:
 *  - Variable names must only contain upper and lower letters, numbers, and underscore.
 *  - They cannot start with a number.
 *  - They cannot be empty {}.
 *  - Two variables cannot be directly touching {like}{this}.
 *  - A variable cannot be directly touching a star {likeThis}* or *{likeThis}.
 *
 *  If variable names don't match these rules, a ValidationError will be thrown.
 */
export declare function parseTemplate(template: string): ParsedTemplate;
export declare let extractTemplateVariablesFromPathUsingRegex: (namedCaptureRegex: string, path: string) => Record<string, string> | null;
export declare let extractTemplateVariablesFromPath: (template: string, path: string) => Record<string, string> | null;
export declare let insertVariablesIntoTemplate: (vars: Record<string, string>, template: string) => string;
/** Given a template string like "/posts/{postId}.json",
 *  query the replica for docs with matching paths.
 *
 * See the docs for matchTemplateAndPath for details on template strings.
 *
 * You can get the variables out of your document paths like this:
 *
 *      let template = '/posts/{postId}.json';
 *      let docs = await queryByTemplate(myReplica, template);
 *      for (let doc of docs) {
 *          // vars will be like { postId: 'abc' }
 *          let vars = extractTemplateVariablesFromPath(template, doc.path);
 *      }
 */
export declare function queryByTemplate<F = DefaultFormats>(replica: Replica, template: string, moreQueryOptions?: Omit<Query<[string]>, "formats">, formats?: FormatsArg<F>): Promise<FormatDocType<F>[]>;
export {};
