import * as dntShim from "../../_dnt.shims.js";
import { AuthorAddress, DocAttachment, DocBase, DocWithAttachment, FormatName, Path, ShareAddress } from "../util/doc-types.js";
import { Query } from "../query/query-types.js";
import { IngestEvent, IReplicaDriver, MultiFormatReplicaOpts, QuerySourceEvent, QuerySourceMode, ReplicaEvent, ReplicaId } from "./replica-types.js";
import { ValidationError } from "../util/errors.js";
import { OrCh } from "../streams/stream_utils.js";
import { DefaultFormat, DefaultFormats, FormatArg, FormatDocType, FormatInputType, FormatsArg } from "../formats/format_types.js";
import { AuthorKeypair } from "../crypto/crypto-types.js";
/**
 * A replica holding a share's data, used to read, write, and synchronise data to.
 *
 * Should be closed using the `close` method when no longer being used.
 *
 * ```ts
 * const gardeningKeypair = await Crypto.generateShareKeypair("gardening");
 *
 * const myReplica = new Replica({
 *  driver: new ReplicaDriverMemory(gardeningKeypair.shareAddress),
 *  config: {
 *    "es.5": { shareSecret: gardeningKeypair.secret }
 *  }
 * });
 * ```
 */
export declare class MultiformatReplica {
    replicaId: ReplicaId;
    /** The address of the share this replica belongs to. */
    share: ShareAddress;
    /** The validator used to validate ingested documents. */
    replicaDriver: IReplicaDriver;
    formatsConfig: Record<string, unknown>;
    private _isClosed;
    private ingestLockStream;
    private eventMultiStream;
    private eventWriter;
    private callbackSink;
    private eraseInterval;
    private expireEventTimeouts;
    constructor({ driver, config }: MultiFormatReplicaOpts);
    /** Returns whether the replica is closed or not. */
    isClosed(): boolean;
    /**
     * Closes the replica, preventing new documents from being ingested or events being emitted.
     * Any methods called after closing will return `ReplicaIsClosedError`.
     * @param erase - Erase the contents of the replica. Defaults to `false`.
     */
    close(erase: boolean): Promise<void>;
    getConfig(key: string): Promise<string | undefined>;
    setConfig(key: string, value: string): Promise<void>;
    listConfigKeys(): Promise<string[]>;
    deleteConfig(key: string): Promise<boolean>;
    /** Returns the max local index of all stored documents */
    getMaxLocalIndex(): Promise<number>;
    /** Returns all documents, including historical versions of documents by other identities. */
    getAllDocs<F = DefaultFormats>(formats?: FormatsArg<F>): Promise<FormatDocType<F>[]>;
    /** Returns latest document from every path. */
    getLatestDocs<F = DefaultFormats>(formats?: FormatsArg<F>): Promise<FormatDocType<F>[]>;
    /** Returns all versions of a document by different authors from a specific path. */
    getAllDocsAtPath<F = DefaultFormats>(path: Path, formats?: FormatsArg<F>): Promise<FormatDocType<F>[]>;
    /** Returns the most recently written version of a document at a path. */
    getLatestDocAtPath<F = DefaultFormat>(path: Path, format?: FormatArg<F>): Promise<FormatDocType<F> | undefined>;
    /** Returns an array of docs for a given query.
      ```
      const myQuery = {
        filter: {
          pathEndsWith: ".txt"
        },
        limit: 5,
      };
  
      const firstFiveTextDocs = await myReplica.queryDocs(myQuery);
      ```
      */
    queryDocs<F = DefaultFormats>(query?: Omit<Query<[string]>, "formats">, formats?: FormatsArg<F>): Promise<FormatDocType<F>[]>;
    /** Returns an array of all unique paths of documents returned by a given query. */
    queryPaths<F = DefaultFormats>(query?: Omit<Query<[string]>, "formats">, formats?: FormatsArg<F>): Promise<Path[]>;
    /** Returns an array of all unique authors of documents returned by a given query. */
    queryAuthors<F = DefaultFormats>(query?: Omit<Query<[string]>, "formats">, formats?: FormatsArg<F>): Promise<AuthorAddress[]>;
    /**
     * Adds a new document to the replica. If a document signed by the same identity exists at the same path, it will be overwritten.
     */
    set<F = DefaultFormat>(keypair: AuthorKeypair, docToSet: Omit<FormatInputType<F>, "format">, format?: FormatArg<F>): Promise<IngestEvent<FormatDocType<F>>>;
    /**
     * Ingest an existing signed document to the replica.
     */
    ingest<F = DefaultFormat>(format: FormatArg<F>, docToIngest: FormatDocType<F>, sourceId: "local" | string): Promise<IngestEvent<FormatDocType<F>>>;
    /**
     * Overwrite every document from this author, including history versions, with an empty doc.
      @returns The number of documents changed, or -1 if there was an error.
     */
    overwriteAllDocsByAuthor<F = DefaultFormat>(keypair: AuthorKeypair, format?: FormatArg<F>): Promise<number | ValidationError>;
    /** Wipe all content from a document at a given path, and erase its attachment (if it has one). */
    wipeDocAtPath<F = DefaultFormat>(keypair: AuthorKeypair, path: string, format?: FormatArg<F>): Promise<IngestEvent<FormatDocType<F>> | ValidationError>;
    private wipeDocument;
    /** Erases expired docs and dangling attachments */
    private pruneExpiredDocsAndAttachments;
    /**
     * Returns a readable stream of replica events, such as new ingestions, document expirations, or the replica preparing to close.
     * @param channel - An optional string representing a channel of events to be subscribed to. Defaults to return all events.
     */
    getEventStream(channel?: OrCh<ReplicaEvent<DocBase<string>>["kind"]>): dntShim.ReadableStream<ReplicaEvent<DocBase<string>>>;
    /**
     * Returns a readable stream of document events which match a given query. The events can represent existing documents, newly ingested documents, or expiring documents.
     */
    getQueryStream<F = DefaultFormats>(query?: Omit<Query<[string]>, "formats">, mode?: QuerySourceMode, formats?: FormatsArg<F>): dntShim.ReadableStream<QuerySourceEvent<FormatDocType<F>>>;
    /**
     * Provide a callback to be triggered every time a replica event occurs.
     * @returns A callback which unsubscribes the event.
     */
    onEvent(callback: (event: ReplicaEvent<DocBase<FormatName>>) => void | Promise<void>): () => void;
    /** Ingest an attachment for a given document.
     *
     * @returns `true` (indicating it was upsert), `false` (indicating this attachment is already in storage), or a `ValidationError` (indicating something went wrong.)
     */
    ingestAttachment<F = DefaultFormat>(format: FormatArg<F>, doc: FormatDocType<F>, attachment: Uint8Array | dntShim.ReadableStream<Uint8Array>, sourceId: "local" | string): Promise<true | false | ValidationError>;
    /** Gets an attachment for a given document. Returns a `ValidationError` if the given document can't have an attachment.
     */
    getAttachment<F = DefaultFormat>(doc: FormatDocType<F>, format?: FormatArg<F>): Promise<DocAttachment | undefined | ValidationError>;
    /** Returns the given array of documents with a new `attachment` property merged in. The value of this property can be:
     * - `DocAttachment`
     * - `undefined` (the associated document can have an attachment, but we don't have a copy)
     * - `ValidationError` (the associated document can't have an attachment)
     */
    addAttachments<F = DefaultFormats>(docs: FormatDocType<F>[], formats?: FormatsArg<F>): Promise<Awaited<DocWithAttachment<FormatDocType<F>>>[]>;
}
