import { AuthorAddress, DocAttachment, DocBase, DocWithAttachment, Path } from "../util/doc-types.js";
import { ValidationError } from "../util/errors.js";
import { Query } from "../query/query-types.js";
import { IngestEvent } from "./replica-types.js";
import { Replica } from "./replica.js";
import { DefaultFormat, DefaultFormats, FormatArg, FormatDocType, FormatInputType, FormatsArg } from "../formats/format_types.js";
import { AuthorKeypair } from "../crypto/crypto-types.js";
/** A cached, synchronous interface to a replica, useful for reactive abstractions. Always returns results from its cache, and proxies the query to the backing replica in case of a cache miss.
 * ```
 * const cache = new ReplicaCache(myReplica);
 * const pngQuery = { filter: { pathEndsWith: ".png" } };
 * let pngDocs = cache.queryDocs(pngQuery);
 * cache.onCacheUpdate(() => {
 *    pngDocs = cache.queryDocs(pngQuery);
 * });
 * ```
 */
export declare class ReplicaCache {
    version: number;
    private replica;
    private docCache;
    private timeToLive;
    private onCacheUpdatedCallbacks;
    private closed;
    private onFireCacheUpdatedsWrapper;
    /**
     * Create a new ReplicaCache.
     * @param timeToLive - The number of milliseconds a cached document is considered valid for.
     * @param onCacheUpdatedWrapper - A function which wraps the firing of all callbacks. Useful for libraries with batching abstractions.
     */
    constructor(replica: Replica, timeToLive?: number, onCacheUpdatedWrapper?: (cb: () => void) => void);
    close(): Promise<void>;
    isClosed(): boolean;
    /** Fetch all versions of all docs from the cache. Returns an empty array in case of a cache miss, and queries the backing replica. */
    getAllDocs<F = DefaultFormats>(formats?: FormatsArg<F>): FormatDocType<F>[];
    /** Fetch latest versions of all docs from the cache. Returns an empty array in case of a cache miss, and queries the backing replica. */
    getLatestDocs<F = DefaultFormats>(formats?: FormatsArg<F>): FormatDocType<F>[];
    /** Fetch all versions of all docs from a certain path from the cache. Returns an empty array in case of a cache miss, and queries the backing replica. */
    getAllDocsAtPath<F = DefaultFormats>(path: Path, formats?: FormatsArg<F>): FormatDocType<F>[];
    /** Fetch latest version of a doc at a path from the cache. Returns an empty array in case of a cache miss, and queries the backing replica. */
    getLatestDocAtPath<F = DefaultFormat>(path: Path, format?: FormatArg<F>): FormatDocType<F> | undefined;
    /** Fetch docs matching a query from the cache. Returns an empty array in case of a cache miss, and queries the backing replica. */
    queryDocs<F = DefaultFormats>(query?: Omit<Query<[string]>, "formats">, formats?: FormatsArg<F>): FormatDocType<F>[];
    /** Returns an array of all unique paths of documents returned by a given query. */
    queryPaths<F = DefaultFormats>(query?: Omit<Query<[string]>, "formats">, formats?: FormatsArg<F>): Path[];
    /** Returns an array of all unique authors of documents returned by a given query. */
    queryAuthors<F = DefaultFormats>(query?: Omit<Query<[string]>, "formats">, formats?: FormatsArg<F>): AuthorAddress[];
    _updateCache(key: string, doc: DocBase<string>): void;
    private fireOnCacheUpdateds;
    /** Subscribes to the cache, calling a callback when previously returned results can be considered stale. Returns a function for unsubscribing. */
    onCacheUpdated(callback: () => void): () => void;
    /** Add a new document directly to the backing replica. */
    set<F = DefaultFormat>(keypair: AuthorKeypair, docToSet: Omit<FormatInputType<F>, "format">, format?: FormatArg<F>): Promise<IngestEvent<FormatDocType<F>> | ValidationError>;
    /** Call this method on the backing replica. */
    overwriteAllDocsByAuthor<F = DefaultFormat>(keypair: AuthorKeypair, format?: FormatArg<F>): Promise<number | ValidationError>;
    wipeDocAtPath<F = DefaultFormat>(keypair: AuthorKeypair, path: string, format?: FormatArg<F>): Promise<IngestEvent<FormatDocType<F>> | ValidationError>;
    private attachmentCache;
    private onReplicaEvent;
    getAttachment<F = DefaultFormat>(doc: FormatDocType<F>, format?: FormatArg<F>): DocAttachment | undefined | ValidationError;
    addAttachments<F = DefaultFormats>(docs: FormatDocType<F>[], formats?: FormatsArg<F>): DocWithAttachment<FormatDocType<F>>[];
}
