import { AuthorKeypair } from "../crypto/crypto-types.js";
import { ShareAddress } from "./doc-types.js";
import { ValidationError } from "./errors.js";
import { Replica } from "../replica/replica.js";
import { Peer } from "../peer/peer.js";
type SharedSettingsOpts = {
    /** A namespace to restrict these settings to. */
    namespace?: string;
    /** Whether to use session storage for these settings. */
    sessionOnly?: true;
};
/** Get and set values from a common pool of settings for Earthstar clients, such as an author, shares, share secrets, and servers.
 *
 * Uses the Storage API, so only clients on the same origin will share the same settings.
 */
export declare class SharedSettings {
    private namespace;
    private storage;
    constructor(opts?: SharedSettingsOpts);
    /** The currently persisted author keypair. */
    get author(): AuthorKeypair | null;
    set author(keypair: AuthorKeypair | null);
    /** An array of shares stored by these settings. */
    get shares(): ShareAddress[];
    /** Add a share to the settings.
     * @returns All stored shares after the addition, or a `ValidationError` if the address is invalid.
     */
    addShare(address: ShareAddress): ValidationError | string[];
    /** Removes a share from settings.
     * @returns All stored shares after the removal, or a `ValidationError` if the share is not yet known.
     */
    removeShare(addressToRemove: string): ValidationError | string[];
    /** A record of known shares and their corresponding secret, if known. */
    get shareSecrets(): Record<string, string>;
    /** Add a secret for a share already known to the settings.
     * @returns The next record of share secret pairs, or returns a `ValidationError` if the share is not known or if the secret is incorrect.
     */
    addSecret(shareAddress: ShareAddress, secret: string): Promise<ValidationError | {
        [x: string]: string;
    }>;
    /** Remove a secret from settings.
     * @returns The next record of share secret pairs, or returns a `ValidationError` if the share is not known.
     */
    removeSecret(shareAddress: ShareAddress): ValidationError | {
        [x: string]: string;
    };
    /** An array of server URLs stored by these settings. */
    get servers(): string[];
    /** Add a server URL to be stored by settings.
     * @returns The list of servers after the addition, or a `ValidationError` if the string is not a valid URL.
     */
    addServer(address: string): string[] | ValidationError;
    /** Remove a server URL from the settings' stored list of servers..
     * @returns The list of servers after the removal, or a `ValidationError` if the string is not yet known.
     */
    removeServer(addressToRemove: string): ValidationError | string[];
    /** Delete all stored settings. */
    clear(): void;
    private authorChangedCbs;
    /** Fires the given callback when the stored author changes. */
    onAuthorChanged(cb: (keypair: AuthorKeypair | null) => void): () => void;
    private sharesChangedCbs;
    /** Fires the given callback when the stored list of shares changes. */
    onSharesChanged(cb: (shares: ShareAddress[]) => void): () => void;
    private shareSecretsChangedCbs;
    /** Fires the given callback when the stored record of share secrets changes. */
    onShareSecretsChanged(cb: (secrets: Record<ShareAddress, string>) => void): () => void;
    private serversChangedCbs;
    /** Fires the given callback when the stored list of server URLs changes. */
    onServersChanged(cb: (shares: string[]) => void): () => void;
    /** Get a new `Peer` preconfigured with shares, secrets, and syncers derived from these settings.
     *
     * When settings are updated, the peer's replicas and syncers will be updated too.
     */
    getPeer({ sync, onCreateReplica }: {
        /** Whether to start syncing using the settings' servers. */
        sync: "once" | "continuous" | false;
        /** Used to create replicas when a new share is added to settings. */
        onCreateReplica: (addr: ShareAddress, secret?: string) => Replica;
    }): {
        /** A preconfigured Peer. */
        peer: Peer;
        /** Stop changes to SharedSettings from propagating to the Peer. */
        unsubscribeFromSettings: () => void;
    };
    private fireAuthorEvent;
    private fireSharesEvent;
    private fireSecretsEvent;
    private fireServersEvent;
    /** Add a new share (and possibly secret) and servers using an Earthstar invitation URL. */
    redeemInvitationURL(url: string): Promise<true | ValidationError>;
}
export {};
